from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any, ClassVar, TypeAlias, TypeVar, final

from returns.functions import identity
from returns.future import FutureResult
from returns.interfaces.specific import reader
from returns.io import IOResult
from returns.primitives.container import BaseContainer
from returns.primitives.hkt import Kind2, SupportsKind2, dekind
from returns.result import Result

if TYPE_CHECKING:
    # We need this condition to make sure Python can solve cycle imports.
    # But, since we only use these values in types, it is not important.
    from returns.context.requires_context_future_result import (
        RequiresContextFutureResult,
    )
    from returns.context.requires_context_ioresult import (
        RequiresContextIOResult,
    )
    from returns.context.requires_context_result import RequiresContextResult

# Context:
_EnvType_contra = TypeVar('_EnvType_contra', contravariant=True)
_NewEnvType = TypeVar('_NewEnvType')
_ReturnType_co = TypeVar('_ReturnType_co', covariant=True)
_NewReturnType = TypeVar('_NewReturnType')

_ValueType = TypeVar('_ValueType')
_ErrorType = TypeVar('_ErrorType')

# Helpers:
_FirstType = TypeVar('_FirstType')

# Type Aliases:
#: Sometimes ``RequiresContext`` and other similar types might be used with
#: no explicit dependencies so we need to have this type alias for Any.
NoDeps = Any


@final
class RequiresContext(  # type: ignore[type-var]
    BaseContainer,
    SupportsKind2['RequiresContext', _ReturnType_co, _EnvType_contra],
    reader.ReaderBased2[_ReturnType_co, _EnvType_contra],
):
    """
    The ``RequiresContext`` container.

    It's main purpose is to wrap some specific function
    and to provide tools to compose other functions around it
    without actually calling it.

    The ``RequiresContext`` container passes the state
    you want to share between functions.
    Functions may read that state, but can't change it.
    The ``RequiresContext`` container
    lets us access shared immutable state within a specific context.

    It can be used for lazy evaluation and typed dependency injection.

    ``RequiresContext`` is used with functions that never fail.
    If you want to use ``RequiresContext`` with returns ``Result``
    then consider using ``RequiresContextResult`` instead.

    Note:
        This container does not wrap ANY value. It wraps only functions.
        You won't be able to supply arbitrary types!

    See also:
        - https://dev.to/gcanti/getting-started-with-fp-ts-reader-1ie5
        - https://en.wikipedia.org/wiki/Lazy_evaluation
        - https://bit.ly/2R8l4WK

    """

    __slots__ = ()

    #: This field has an extra 'RequiresContext' just because `mypy` needs it.
    _inner_value: Callable[[_EnvType_contra], _ReturnType_co]

    #: A convenient placeholder to call methods created by `.from_value()`:
    no_args: ClassVar[NoDeps] = object()

    def __init__(
        self,
        inner_value: Callable[[_EnvType_contra], _ReturnType_co],
    ) -> None:
        """
        Public constructor for this type. Also required for typing.

        Only allows functions of kind ``* -> *``.

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> str(RequiresContext(lambda deps: deps + 1))
          '<RequiresContext: <function <lambda> at ...>>'

        """
        super().__init__(inner_value)

    def __call__(self, deps: _EnvType_contra) -> _ReturnType_co:
        """
        Evaluates the wrapped function.

        .. code:: python

          >>> from returns.context import RequiresContext

          >>> def first(lg: bool) -> RequiresContext[int, float]:
          ...     # `deps` has `float` type here:
          ...     return RequiresContext(
          ...         lambda deps: deps if lg else -deps,
          ...     )

          >>> instance = first(False)  # creating `RequiresContext` instance
          >>> assert instance(3.5) == -3.5 # calling it with `__call__`

          >>> # Example with another logic:
          >>> assert first(True)(3.5) == 3.5

        In other things, it is a regular python magic method.
        """
        return self._inner_value(deps)

    def map(
        self,
        function: Callable[[_ReturnType_co], _NewReturnType],
    ) -> RequiresContext[_NewReturnType, _EnvType_contra]:
        """
        Allows to compose functions inside the wrapped container.

        Here's how it works:

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> def first(lg: bool) -> RequiresContext[int, float]:
          ...     # `deps` has `float` type here:
          ...     return RequiresContext(
          ...         lambda deps: deps if lg else -deps,
          ...     )

          >>> assert first(True).map(lambda number: number * 10)(2.5) == 25.0
          >>> assert first(False).map(lambda number: number * 10)(0.1) -1.0

        """
        return RequiresContext(lambda deps: function(self(deps)))

    def apply(
        self,
        container: Kind2[
            RequiresContext,
            Callable[[_ReturnType_co], _NewReturnType],
            _EnvType_contra,
        ],
    ) -> RequiresContext[_NewReturnType, _EnvType_contra]:
        """
        Calls a wrapped function in a container on this container.

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> assert RequiresContext.from_value('a').apply(
          ...    RequiresContext.from_value(lambda inner: inner + 'b')
          ... )(...) == 'ab'

        """
        return RequiresContext(
            lambda deps: self.map(dekind(container)(deps))(deps),
        )

    def bind(
        self,
        function: Callable[
            [_ReturnType_co],
            Kind2[RequiresContext, _NewReturnType, _EnvType_contra],
        ],
    ) -> RequiresContext[_NewReturnType, _EnvType_contra]:
        """
        Composes a container with a function returning another container.

        This is useful when you do several computations that rely on the
        same context.

        .. code:: python

          >>> from returns.context import RequiresContext

          >>> def first(lg: bool) -> RequiresContext[int, float]:
          ...     # `deps` has `float` type here:
          ...     return RequiresContext(
          ...         lambda deps: deps if lg else -deps,
          ...     )

          >>> def second(number: int) -> RequiresContext[str, float]:
          ...     # `deps` has `float` type here:
          ...     return RequiresContext(
          ...         lambda deps: '>=' if number >= deps else '<',
          ...     )

          >>> assert first(True).bind(second)(1) == '>='
          >>> assert first(False).bind(second)(2) == '<'

        """
        return RequiresContext(lambda deps: dekind(function(self(deps)))(deps))

    #: Alias for `bind_context` method, it is the same as `bind` here.
    bind_context = bind

    def modify_env(
        self,
        function: Callable[[_NewEnvType], _EnvType_contra],
    ) -> RequiresContext[_ReturnType_co, _NewEnvType]:
        """
        Allows to modify the environment type.

        .. code:: python

          >>> from returns.context import RequiresContext

          >>> def mul(arg: int) -> RequiresContext[float, int]:
          ...     return RequiresContext(lambda deps: arg * deps)

          >>> assert mul(3).modify_env(int)('2') == 6

        """
        return RequiresContext(lambda deps: self(function(deps)))

    @classmethod
    def ask(cls) -> RequiresContext[_EnvType_contra, _EnvType_contra]:
        """
        Get current context to use the dependencies.

        It is a common scenario when you need to use the environment.
        For example, you want to do some context-related computation,
        but you don't have the context instance at your disposal.
        That's where ``.ask()`` becomes useful!

        .. code:: python

          >>> from typing_extensions import TypedDict
          >>> class Deps(TypedDict):
          ...     message: str

          >>> def first(lg: bool) -> RequiresContext[int, Deps]:
          ...     # `deps` has `Deps` type here:
          ...     return RequiresContext(
          ...         lambda deps: deps['message'] if lg else 'error',
          ...     )

          >>> def second(text: str) -> RequiresContext[int, Deps]:
          ...     return first(len(text) > 3)

          >>> assert second('abc')({'message': 'ok'}) == 'error'
          >>> assert second('abcd')({'message': 'ok'}) == 'ok'

        And now imagine that you have to change this ``3`` limit.
        And you want to be able to set it via environment as well.
        Ok, let's fix it with the power of ``RequiresContext.ask()``!

        .. code:: python

          >>> from typing_extensions import TypedDict
          >>> class Deps(TypedDict):
          ...     message: str
          ...     limit: int   # note this new field!

          >>> def new_first(lg: bool) -> RequiresContext[int, Deps]:
          ...     # `deps` has `Deps` type here:
          ...     return RequiresContext(
          ...         lambda deps: deps['message'] if lg else 'err',
          ...     )

          >>> def new_second(text: str) -> RequiresContext[int, Deps]:
          ...     return RequiresContext[int, Deps].ask().bind(
          ...         lambda deps: new_first(len(text) > deps.get('limit', 3)),
          ...     )

          >>> assert new_second('abc')({'message': 'ok', 'limit': 2}) == 'ok'
          >>> assert new_second('abcd')({'message': 'ok'}) == 'ok'
          >>> assert new_second('abcd')({'message': 'ok', 'limit': 5}) == 'err'

        That's how ``ask`` works.

        This class contains methods that require
        to explicitly set type annotations. Why?
        Because it is impossible to figure out the type without them.
        So, here's how you should use them:

        .. code:: python

            RequiresContext[int, Dict[str, str]].ask()

        Otherwise, your ``.ask()`` method
        will return ``RequiresContext[Never, Never]``,
        which is unusable:

        .. code:: python

            env = RequiresContext.ask()
            env(some_deps)

        And ``mypy`` will warn you: ``error: Need type annotation for '...'``

        See also:
            - https://dev.to/gcanti/getting-started-with-fp-ts-reader-1ie5

        """
        return RequiresContext(identity)

    @classmethod
    def from_value(
        cls,
        inner_value: _FirstType,
    ) -> RequiresContext[_FirstType, NoDeps]:
        """
        Used to return some specific value from the container.

        Consider this method as some kind of factory.
        Passed value will be a return type.
        Make sure to use :attr:`~RequiresContext.no_args`
        for getting the unit value.

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> unit = RequiresContext.from_value(5)
          >>> assert unit(RequiresContext.no_args) == 5

        Might be used with or without direct type hint.
        """
        return RequiresContext(lambda _: inner_value)

    @classmethod
    def from_context(
        cls,
        inner_value: RequiresContext[_NewReturnType, _NewEnvType],
    ) -> RequiresContext[_NewReturnType, _NewEnvType]:
        """
        Used to create new containers from existing ones.

        Used as a part of ``ReaderBased2`` interface.

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> unit = RequiresContext.from_value(5)
          >>> assert RequiresContext.from_context(unit)(...) == unit(...)

        """
        return inner_value

    @classmethod
    def from_requires_context_result(
        cls,
        inner_value: RequiresContextResult[
            _ValueType, _ErrorType, _EnvType_contra
        ],
    ) -> RequiresContext[Result[_ValueType, _ErrorType], _EnvType_contra]:
        """
        Typecasts ``RequiresContextResult`` to ``RequiresContext`` instance.

        Breaks ``RequiresContextResult[a, b, e]``
        into ``RequiresContext[Result[a, b], e]``.

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> from returns.context import RequiresContextResult
          >>> from returns.result import Success
          >>> assert RequiresContext.from_requires_context_result(
          ...    RequiresContextResult.from_value(1),
          ... )(...) == Success(1)

        Can be reverted with ``RequiresContextResult.from_typecast``.

        """
        return RequiresContext(inner_value)

    @classmethod
    def from_requires_context_ioresult(
        cls,
        inner_value: RequiresContextIOResult[
            _ValueType, _ErrorType, _EnvType_contra
        ],
    ) -> RequiresContext[IOResult[_ValueType, _ErrorType], _EnvType_contra]:
        """
        Typecasts ``RequiresContextIOResult`` to ``RequiresContext`` instance.

        Breaks ``RequiresContextIOResult[a, b, e]``
        into ``RequiresContext[IOResult[a, b], e]``.

        .. code:: python

          >>> from returns.context import RequiresContext
          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess
          >>> assert RequiresContext.from_requires_context_ioresult(
          ...    RequiresContextIOResult.from_value(1),
          ... )(...) == IOSuccess(1)

        Can be reverted with ``RequiresContextIOResult.from_typecast``.

        """
        return RequiresContext(inner_value)

    @classmethod
    def from_requires_context_future_result(
        cls,
        inner_value: RequiresContextFutureResult[
            _ValueType,
            _ErrorType,
            _EnvType_contra,
        ],
    ) -> RequiresContext[FutureResult[_ValueType, _ErrorType], _EnvType_contra]:
        """
        Typecasts ``RequiresContextIOResult`` to ``RequiresContext`` instance.

        Breaks ``RequiresContextIOResult[a, b, e]``
        into ``RequiresContext[IOResult[a, b], e]``.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContext
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess

          >>> container = RequiresContext.from_requires_context_future_result(
          ...    RequiresContextFutureResult.from_value(1),
          ... )
          >>> assert anyio.run(
          ...     container, RequiresContext.no_args,
          ... ) == IOSuccess(1)

        Can be reverted with ``RequiresContextFutureResult.from_typecast``.

        """
        return RequiresContext(inner_value)


# Aliases

#: Sometimes `RequiresContext` is too long to type.
Reader: TypeAlias = RequiresContext



from returns.io import IO


def test_io_pickle():
    """Tests how pickle protocol works for containers."""
    assert IO(1).__getstate__() == {'container_value': 1}


def test_io_pickle_restore():
    """Ensures that object can be restored."""
    container = IO(2)
    container.__setstate__({'container_value': 1})
    assert container == IO(1)



import pytest

from returns.future import Future


def test_nonequality():
    """Ensures that containers can be compared."""
    assert Future.from_value(1) != Future.from_value(1)
    assert hash(Future.from_value(1))


@pytest.mark.anyio
async def test_equality():
    """Ensures that containers are not compared to regular values."""
    assert await Future.from_value(2) == await Future.from_value(2)



from itertools import starmap
from typing import NamedTuple, final

from mypy.nodes import ArgKind, Context, TempNode
from mypy.types import CallableType
from mypy.types import Type as MypyType


class _FuncArgStruct(NamedTuple):
    """Basic struct to represent function arguments."""

    name: str | None
    type: MypyType  # noqa: WPS125
    kind: ArgKind


@final
class FuncArg(_FuncArgStruct):
    """Representation of function arg with all required fields and methods."""

    def expression(self, context: Context) -> TempNode:
        """Hack to pass unexisting `Expression` to typechecker."""
        return TempNode(self.type, context=context)

    @classmethod
    def from_callable(cls, function_def: CallableType) -> list['FuncArg']:
        """Public constructor to create FuncArg lists from callables."""
        parts = zip(
            function_def.arg_names,
            function_def.arg_types,
            function_def.arg_kinds,
            strict=False,
        )
        return list(starmap(cls, parts))



import anyio
import pytest

from returns.future import Future, future
from returns.io import IO


@future
async def _coro(arg: int) -> float:
    assert isinstance(arg, int)
    await anyio.sleep(0)
    return arg / 2


@pytest.mark.anyio
async def test_safe_decorator():
    """Ensure that coroutine marked with ``@future`` returns ``Future``."""
    future_instance = _coro(1)

    assert isinstance(future_instance, Future)
    assert await future_instance == IO(0.5)



from typing import TypeVar

from returns.io import IO

_ValueType = TypeVar('_ValueType')


def unsafe_perform_io(wrapped_in_io: IO[_ValueType]) -> _ValueType:
    """
    Compatibility utility and escape mechanism from ``IO`` world.

    Just unwraps the internal value
    from :class:`returns.io.IO` container.
    Should be used with caution!
    Since it might be overused by lazy and ignorant developers.

    It is recommended to have only one place (module / file)
    in your program where you allow unsafe operations.

    We recommend to use ``import-linter`` to enforce this rule.

    .. code:: python

      >>> from returns.io import IO
      >>> assert unsafe_perform_io(IO(1)) == 1

    See also:
        - https://github.com/seddonym/import-linter

    """
    return wrapped_in_io._inner_value  # noqa: SLF001



from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Generic, Never, Protocol, TypeVar

from typing_extensions import TypeVarTuple

_InstanceType_co = TypeVar('_InstanceType_co', covariant=True)
_TypeArgType1_co = TypeVar('_TypeArgType1_co', covariant=True)
_TypeArgType2_co = TypeVar('_TypeArgType2_co', covariant=True)
_TypeArgType3_co = TypeVar('_TypeArgType3_co', covariant=True)

_FunctionDefType_co = TypeVar(
    '_FunctionDefType_co',
    bound=Callable,
    covariant=True,  # This is a must! Otherwise it would not work.
)
_FunctionType = TypeVar(
    '_FunctionType',
    bound=Callable,
)

_UpdatedType = TypeVar('_UpdatedType')
_TypeVars = TypeVarTuple('_TypeVars')


class KindN(Generic[_InstanceType_co, *_TypeVars]):
    """
    Emulation support for Higher Kinded Types.

    Consider ``KindN`` to be an alias of ``Generic`` type.
    But with some extra goodies.

    ``KindN`` is the top-most type for other ``Kind`` types
    like ``Kind1``, ``Kind2``, ``Kind3``, etc.

    The only difference between them is how many type arguments they can hold.
    ``Kind1`` can hold just two type arguments: ``Kind1[IO, int]``
    which is almost equals to ``IO[int]``.
    ``Kind2`` can hold just two type arguments: ``Kind2[IOResult, int, str]``
    which is almost equals to ``IOResult[int, str]``.
    And so on.

    The idea behind ``KindN`` is that one cannot write this code:

    .. code:: python

      from typing import TypeVar

      T = TypeVar('T')
      V = TypeVar('V')

      def impossible(generic: T, value: V) -> T[V]:
          return generic(value)

    But, with ``KindN`` this becomes possible in a form of ``Kind1[T, V]``.

    .. note::
        To make sure it works correctly,
        your type has to be a subtype of ``KindN``.

    We use a custom ``mypy`` plugin to make sure types are correct.
    Otherwise, it is currently impossible to properly type this.

    We use "emulated Higher Kinded Types" concept.
    Read the whitepaper: https://bit.ly/2ABACx2

    ``KindN`` does not exist in runtime. It is used just for typing.
    There are (and must be) no instances of this type directly.

    .. rubric:: Implementation details

    We didn't use ``ABCMeta`` to disallow its creation,
    because we don't want to have
    a possible metaclass conflict with other metaclasses.
    Current API allows you to mix ``KindN`` anywhere.

    We allow ``_InstanceType_co`` of ``KindN``
    to be ``Instance`` type or ``TypeVarType`` with ``bound=...``.

    See also:
        - https://arrow-kt.io/docs/0.10/patterns/glossary/#higher-kinds
        - https://github.com/gcanti/fp-ts/blob/master/docs/guides/HKT.md
        - https://bow-swift.io/docs/fp-concepts/higher-kinded-types
        - https://github.com/pelotom/hkts

    """

    __slots__ = ()

    if TYPE_CHECKING:  # noqa: WPS604 # pragma: no cover

        def __getattr__(self, attrname: str):
            """
            This function is required for ``get_attribute_hook`` in mypy plugin.

            It is never called in real-life, because ``KindN`` is abstract.
            It only exists during the type-checking phase.
            """


#: Type alias for kinds with one type argument.
Kind1 = KindN[_InstanceType_co, _TypeArgType1_co, Any, Any]

#: Type alias for kinds with two type arguments.
Kind2 = KindN[_InstanceType_co, _TypeArgType1_co, _TypeArgType2_co, Any]

#: Type alias for kinds with three type arguments.
Kind3 = KindN[
    _InstanceType_co, _TypeArgType1_co, _TypeArgType2_co, _TypeArgType3_co
]


class SupportsKindN(KindN[_InstanceType_co, *_TypeVars]):
    """
    Base class for your containers.

    Notice, that we use ``KindN`` / ``Kind1`` to annotate values,
    but we use ``SupportsKindN`` / ``SupportsKind1`` to inherit from.

    .. rubric:: Implementation details

    The only thing this class does is: making sure that the resulting classes
    won't have ``__getattr__`` available during the typechecking phase.

    Needless to say, that ``__getattr__`` during runtime - never exists at all.
    """

    __slots__ = ()

    __getattr__: None  # type: ignore


#: Type alias used for inheritance with one type argument.
SupportsKind1 = SupportsKindN[
    _InstanceType_co,
    _TypeArgType1_co,
    Never,
    Never,
]

#: Type alias used for inheritance with two type arguments.
SupportsKind2 = SupportsKindN[
    _InstanceType_co,
    _TypeArgType1_co,
    _TypeArgType2_co,
    Never,
]

#: Type alias used for inheritance with three type arguments.
SupportsKind3 = SupportsKindN[
    _InstanceType_co,
    _TypeArgType1_co,
    _TypeArgType2_co,
    _TypeArgType3_co,
]


def dekind(
    kind: KindN[
        _InstanceType_co, _TypeArgType1_co, _TypeArgType2_co, _TypeArgType3_co
    ],
) -> _InstanceType_co:
    """
    Turns ``Kind1[IO, int]`` type into real ``IO[int]`` type.

    Should be used when you are left with accidental ``KindN`` instance
    when you really want to have the real type.

    Works with type arguments of any length.

    We use a custom ``mypy`` plugin to make sure types are correct.
    Otherwise, it is currently impossible to properly type this.

    In runtime it just returns the passed argument, nothing really happens:

    .. code:: python

      >>> from returns.io import IO
      >>> from returns.primitives.hkt import Kind1

      >>> container: Kind1[IO, int] = IO(1)
      >>> assert dekind(container) is container

    However, please, do not use this function
    unless you know exactly what you are doing and why do you need it.
    """
    return kind  # type: ignore


# Utils to define kinded functions
# ================================


# TODO: in the future we would be able to write a custom plugin
# with `transform_kind(T) -> T'` support.
# It would visit all the possible `KindN[]` types in any type and run `dekind`
# on them, so this will be how it works:
# in:  => Callable[[KindN[IO[Any], int]], KindN[IO[Any], str]]
# out: => Callable[[IO[int]], IO[str]]
# This will allow to have better support for callable protocols and similar.
# Blocked by: https://github.com/python/mypy/issues/9001
class Kinded(Protocol[_FunctionDefType_co]):  # type: ignore
    """
    Protocol that tracks kinded functions calls.

    We use a custom ``mypy`` plugin to make sure types are correct.
    Otherwise, it is currently impossible to properly type this.
    """

    __slots__ = ()

    #: Used to translate `KindN` into real types.
    __call__: _FunctionDefType_co

    def __get__(
        self,
        instance: _UpdatedType,
        type_,
    ) -> Callable[..., _UpdatedType]:
        """Used to decorate and properly analyze method calls."""


def kinded(function: _FunctionType) -> Kinded[_FunctionType]:
    """
    Decorator to be used when you want to dekind the function's return type.

    Does nothing in runtime, just returns its argument.

    We use a custom ``mypy`` plugin to make sure types are correct.
    Otherwise, it is currently impossible to properly type this.

    Here's an example of how it should be used:

    .. code:: python

      >>> from typing import TypeVar
      >>> from returns.primitives.hkt import KindN, kinded
      >>> from returns.interfaces.bindable import BindableN

      >>> _Binds = TypeVar('_Binds', bound=BindableN)  # just an example
      >>> _Type1 = TypeVar('_Type1')
      >>> _Type2 = TypeVar('_Type2')
      >>> _Type3 = TypeVar('_Type3')

      >>> @kinded
      ... def bindable_identity(
      ...    container: KindN[_Binds, _Type1, _Type2, _Type3],
      ... ) -> KindN[_Binds, _Type1, _Type2, _Type3]:
      ...     return container  # just do nothing

    As you can see, here we annotate our return type as
    ``-> KindN[_Binds, _Type1, _Type2, _Type3]``,
    it would be true without ``@kinded`` decorator.

    But, ``@kinded`` decorator dekinds the return type and infers
    the real type behind it:

    .. code:: python

      >>> from returns.io import IO, IOResult

      >>> assert bindable_identity(IO(1)) == IO(1)
      >>> # => Revealed type: 'IO[int]'

      >>> iores: IOResult[int, str] = IOResult.from_value(1)
      >>> assert bindable_identity(iores) == iores
      >>> # => Revealed type: 'IOResult[int, str]'

    The difference is very clear in ``methods`` modules, like:

    - Raw :func:`returns.methods.bind.internal_bind`
      that returns ``KindN`` instance
    - User-facing :func:`returns.methods.bind.bind`
      that returns the container type

    You must use this decorator for your own kinded functions as well.
    """
    return function  # type: ignore



from collections.abc import Iterable

from mypy.typeops import erase_to_bound
from mypy.types import (
    AnyType,
    CallableType,
    DeletedType,
    ErasedType,
    Instance,
    LiteralType,
    NoneType,
    Overloaded,
    PartialType,
    ProperType,
    TupleType,
    TypedDictType,
    TypeOfAny,
    TypeType,
    TypeVarType,
    UnboundType,
    UninhabitedType,
    UnionType,
    get_proper_type,
)
from mypy.types import Type as MypyType

from returns.contrib.mypy._consts import TYPED_KINDN

# TODO: replace with real `TypeTranslator` in the next mypy release.
_LEAF_TYPES = (
    UnboundType,
    AnyType,
    NoneType,
    UninhabitedType,
    ErasedType,
    DeletedType,
    TypeVarType,
    PartialType,
)


def translate_kind_instance(typ: MypyType) -> ProperType:  # noqa: C901, WPS210, WPS212, WPS231
    """
    We use this ugly hack to translate ``KindN[x, y]`` into ``x[y]``.

    This is required due to the fact that ``KindN``
    can be nested in other types, like: ``List[KindN[...]]``.

    We will refactor this code after ``TypeTranslator``
    is released in ``mypy@0.800`` version.
    """
    typ = get_proper_type(typ)

    if isinstance(typ, _LEAF_TYPES):  # noqa: WPS223
        return typ
    if isinstance(typ, Instance):
        last_known_value: LiteralType | None = None
        if typ.last_known_value is not None:
            raw_last_known_value = translate_kind_instance(typ.last_known_value)
            assert isinstance(raw_last_known_value, LiteralType)
            last_known_value = raw_last_known_value
        instance = Instance(
            typ=typ.type,
            args=_translate_types(typ.args),
            line=typ.line,
            column=typ.column,
            last_known_value=last_known_value,
        )
        if typ.type.fullname == TYPED_KINDN:  # That's where we do the change
            return _process_kinded_type(instance)
        return instance

    if isinstance(typ, CallableType):
        return typ.copy_modified(
            arg_types=_translate_types(typ.arg_types),
            ret_type=translate_kind_instance(typ.ret_type),
        )
    if isinstance(typ, TupleType):
        return TupleType(
            _translate_types(typ.items),
            translate_kind_instance(typ.partial_fallback),  # type: ignore
            typ.line,
            typ.column,
        )
    if isinstance(typ, TypedDictType):
        dict_items: dict[str, MypyType] = {
            item_name: translate_kind_instance(item_type)
            for item_name, item_type in typ.items.items()
        }
        return TypedDictType(
            dict_items,
            required_keys=typ.required_keys,
            readonly_keys=typ.readonly_keys,
            fallback=translate_kind_instance(typ.fallback),  # type: ignore
            line=typ.line,
            column=typ.column,
        )
    if isinstance(typ, LiteralType):
        fallback = translate_kind_instance(typ.fallback)
        assert isinstance(fallback, Instance)
        return LiteralType(
            value=typ.value,
            fallback=fallback,
            line=typ.line,
            column=typ.column,
        )
    if isinstance(typ, UnionType):
        return UnionType(_translate_types(typ.items), typ.line, typ.column)
    if isinstance(typ, Overloaded):
        functions: list[CallableType] = []
        for func in typ.items:
            new = translate_kind_instance(func)
            assert isinstance(new, CallableType)
            functions.append(new)
        return Overloaded(items=functions)
    if isinstance(typ, TypeType):
        return TypeType.make_normalized(
            translate_kind_instance(typ.item),
            line=typ.line,
            column=typ.column,
        )
    return typ


def _translate_types(types: Iterable[MypyType]) -> list[MypyType]:
    return [translate_kind_instance(typ) for typ in types]


def _process_kinded_type(kind: Instance) -> ProperType:
    """Recursively process all type arguments in a kind."""
    if not kind.args:
        return kind

    real_type = get_proper_type(kind.args[0])
    if isinstance(real_type, TypeVarType):
        return get_proper_type(erase_to_bound(real_type))
    if isinstance(real_type, Instance):
        return real_type.copy_modified(
            args=kind.args[1 : len(real_type.args) + 1],
        )

    # This should never happen, probably can be an exception:
    return AnyType(TypeOfAny.implementation_artifact)



import pytest

from returns.result import Success, safe


@safe
def _function(number: int) -> float:
    return number / number


@safe(exceptions=(ZeroDivisionError,))
def _function_two(number: int | str) -> float:
    assert isinstance(number, int)
    return number / number


@safe((ZeroDivisionError,))  # no name
def _function_three(number: int | str) -> float:
    assert isinstance(number, int)
    return number / number


def test_safe_success():
    """Ensures that safe decorator works correctly for Success case."""
    assert _function(1) == Success(1.0)


def test_safe_failure():
    """Ensures that safe decorator works correctly for Failure case."""
    failed = _function(0)
    assert isinstance(failed.failure(), ZeroDivisionError)


def test_safe_failure_with_expected_error():
    """Ensures that safe decorator works correctly for Failure case."""
    failed = _function_two(0)
    assert isinstance(failed.failure(), ZeroDivisionError)

    failed2 = _function_three(0)
    assert isinstance(failed2.failure(), ZeroDivisionError)


def test_safe_failure_with_non_expected_error():
    """Ensures that safe decorator works correctly for Failure case."""
    with pytest.raises(AssertionError):
        _function_two('0')



"""
An interface that represents a pure computation result.

For impure result see
:class:`returns.interfaces.specific.ioresult.IOResultLikeN` type.
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable
from typing import TYPE_CHECKING, Never, TypeVar

from returns.interfaces import equable, failable, unwrappable
from returns.primitives.hkt import KindN

if TYPE_CHECKING:
    from returns.result import Result  # noqa: WPS433

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_ResultLikeType = TypeVar('_ResultLikeType', bound='ResultLikeN')

# New values:
_ValueType = TypeVar('_ValueType')
_ErrorType = TypeVar('_ErrorType')

# Unwrappable:
_FirstUnwrappableType = TypeVar('_FirstUnwrappableType')
_SecondUnwrappableType = TypeVar('_SecondUnwrappableType')


class ResultLikeN(
    failable.DiverseFailableN[_FirstType, _SecondType, _ThirdType],
):
    """
    Base types for types that looks like ``Result`` but cannot be unwrapped.

    Like ``RequiresContextResult`` or ``FutureResult``.
    """

    __slots__ = ()

    @abstractmethod
    def bind_result(
        self: _ResultLikeType,
        function: Callable[[_FirstType], Result[_UpdatedType, _SecondType]],
    ) -> KindN[_ResultLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Runs ``Result`` returning function over a container."""

    @classmethod
    @abstractmethod
    def from_result(
        cls: type[_ResultLikeType],
        inner_value: Result[_ValueType, _ErrorType],
    ) -> KindN[_ResultLikeType, _ValueType, _ErrorType, _ThirdType]:
        """Unit method to create new containers from any raw value."""


#: Type alias for kinds with two type arguments.
ResultLike2 = ResultLikeN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
ResultLike3 = ResultLikeN[_FirstType, _SecondType, _ThirdType]


class UnwrappableResult(
    ResultLikeN[_FirstType, _SecondType, _ThirdType],
    unwrappable.Unwrappable[_FirstUnwrappableType, _SecondUnwrappableType],
    equable.Equable,
):
    """
    Intermediate type with 5 type arguments that represents unwrappable result.

    It is a raw type and should not be used directly.
    Use ``ResultBasedN`` and ``IOResultBasedN`` instead.
    """

    __slots__ = ()


class ResultBasedN(
    UnwrappableResult[
        _FirstType,
        _SecondType,
        _ThirdType,
        # Unwraps:
        _FirstType,
        _SecondType,
    ],
):
    """
    Base type for real ``Result`` types.

    Can be unwrapped.
    """

    __slots__ = ()


#: Type alias for kinds with two type arguments.
ResultBased2 = ResultBasedN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
ResultBased3 = ResultBasedN[_FirstType, _SecondType, _ThirdType]



from returns.io import IOFailure, IOResult, IOSuccess
from returns.result import Success

container: IOResult[int, str] = IOSuccess(42)
match container:
    # Matches if the result stored inside `IOSuccess` is `42`
    # We need to use `Success` until the custom matching protocol
    # is released. For more information, please visit:
    # https://www.python.org/dev/peps/pep-0622/#custom-matching-protocol
    case IOSuccess(Success(42)):
        print('Result is "42"')

    # Matches any `IOSuccess` instance
    # and binds its value to the `value` variable
    case IOSuccess(value):
        print(f'Result is "{value}"')

    # Matches any `IOFailure` instance
    case IOFailure(_):
        print('A failure was occurred')



from returns.context import RequiresContext


def test_context_ask():
    """Ensures that ``ask`` method works correctly."""
    assert RequiresContext[str, int].ask()(1) == 1
    assert RequiresContext[int, str].ask()('a') == 'a'


def test_requires_context_from_value():
    """Ensures that ``from_value`` method works correctly."""
    assert RequiresContext.from_value(1)(RequiresContext.no_args) == 1
    assert RequiresContext.from_value(2)(1) == 2



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, TypeVar

from returns.interfaces.specific.io import IOLikeN
from returns.primitives.hkt import Kinded, KindN, kinded

if TYPE_CHECKING:
    from returns.io import IO  # noqa: WPS433

_FirstType_contra = TypeVar('_FirstType_contra', contravariant=True)
_SecondType = TypeVar('_SecondType')
_ThirdType_contra = TypeVar('_ThirdType_contra', contravariant=True)
_UpdatedType_co = TypeVar('_UpdatedType_co', covariant=True)

_IOLikeKind = TypeVar('_IOLikeKind', bound=IOLikeN)


def bind_io(
    function: Callable[[_FirstType_contra], IO[_UpdatedType_co]],
) -> Kinded[
    Callable[
        [KindN[_IOLikeKind, _FirstType_contra, _SecondType, _ThirdType_contra]],
        KindN[_IOLikeKind, _UpdatedType_co, _SecondType, _ThirdType_contra],
    ]
]:
    """
    Composes successful container with a function that returns a container.

    In other words, it modifies the function's
    signature from:
    ``a -> IO[b]``
    to:
    ``Container[a, c] -> Container[b, c]``

    .. code:: python

      >>> from returns.io import IOSuccess, IOFailure
      >>> from returns.io import IO
      >>> from returns.pointfree import bind_io

      >>> def returns_io(arg: int) -> IO[int]:
      ...     return IO(arg + 1)

      >>> bound = bind_io(returns_io)
      >>> assert bound(IO(1)) == IO(2)
      >>> assert bound(IOSuccess(1)) == IOSuccess(2)
      >>> assert bound(IOFailure(1)) == IOFailure(1)

    """

    @kinded
    def factory(
        container: KindN[
            _IOLikeKind, _FirstType_contra, _SecondType, _ThirdType_contra
        ],
    ) -> KindN[_IOLikeKind, _UpdatedType_co, _SecondType, _ThirdType_contra]:
        return container.bind_io(function)

    return factory



from collections.abc import Callable
from typing import TypeVar

from returns.future import Future
from returns.interfaces.specific.future import FutureLikeN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_FutureKind = TypeVar('_FutureKind', bound=FutureLikeN)


def bind_future(
    function: Callable[[_FirstType], Future[_UpdatedType]],
) -> Kinded[
    Callable[
        [KindN[_FutureKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_FutureKind, _UpdatedType, _SecondType, _ThirdType],
    ]
]:
    """
    Compose a container and sync function returning ``Future``.

    In other words, it modifies the function
    signature from:
    ``a -> Future[b]``
    to:
    ``Container[a] -> Container[b]``

    Similar to :func:`returns.pointfree.lash`,
    but works for successful containers.
    This is how it should be used:

    .. code:: python

      >>> import anyio
      >>> from returns.pointfree import bind_future
      >>> from returns.future import Future
      >>> from returns.io import IO

      >>> def example(argument: int) -> Future[int]:
      ...     return Future.from_value(argument + 1)

      >>> assert anyio.run(
      ...     bind_future(example)(Future.from_value(1)).awaitable,
      ... ) == IO(2)

    Note, that this function works
    for all containers with ``.bind_future`` method.
    See :class:`returns.primitives.interfaces.specific.future.FutureLikeN`
    for more info.

    """

    @kinded
    def factory(
        container: KindN[_FutureKind, _FirstType, _SecondType, _ThirdType],
    ) -> KindN[_FutureKind, _UpdatedType, _SecondType, _ThirdType]:
        return container.bind_future(function)

    return factory



from returns.maybe import Nothing, Some, maybe


@maybe
def _function(hashmap: dict[str, str], key: str) -> str | None:
    return hashmap.get(key)


def test_maybe_some():
    """Ensures that maybe decorator works correctly for some case."""
    assert _function({'a': 'b'}, 'a') == Some('b')


def test_maybe_nothing():
    """Ensures that maybe decorator works correctly for nothing case."""
    assert _function({'a': 'b'}, 'c') == Nothing



import sys
from collections.abc import Callable, Iterator

import pytest

from returns.trampolines import Trampoline, trampoline


@trampoline
def _accumulate(
    numbers: Iterator[int],
    acc: int = 0,
) -> int | Trampoline[int]:
    number = next(numbers, None)
    if number is None:
        return acc
    return Trampoline(_accumulate, numbers, acc + number)


@trampoline
def _with_func_kwarg(
    numbers: Iterator[int],
    func: int = 0,  # we need this name to match `Trampoline` constructor
) -> int | Trampoline[int]:
    number = next(numbers, None)
    if number is None:
        return func
    return Trampoline(_with_func_kwarg, numbers, func=func + number)


@pytest.mark.parametrize(
    'trampoline_func',
    [
        _accumulate,
        _with_func_kwarg,
    ],
)
@pytest.mark.parametrize(
    'given_range',
    [
        range(0),
        range(1),
        range(2),
        range(5),
        range(sys.getrecursionlimit()),
        range(sys.getrecursionlimit() + 1),
    ],
)
def test_recursion_limit(
    trampoline_func: Callable[[Iterator[int]], int],
    given_range: range,
) -> None:
    """Test that accumulation is correct and no ``RecursionError`` happens."""
    accumulated = trampoline_func(iter(given_range))

    assert accumulated == sum(given_range)



from typing import Final

from mypy.maptype import map_instance_to_supertype
from mypy.nodes import Expression, GeneratorExpr, TypeInfo
from mypy.plugin import MethodContext
from mypy.subtypes import is_subtype
from mypy.typeops import make_simplified_union
from mypy.types import (
    AnyType,
    CallableType,
    Instance,
    TypeOfAny,
    UnionType,
    get_proper_type,
)
from mypy.types import Type as MypyType

_INVALID_DO_NOTATION_SOURCE: Final = (
    'Invalid type supplied in do-notation: expected "{0}", got "{1}"'
)
_LITERAL_GENERATOR_EXPR_REQUIRED: Final = (
    'Literal generator expression is required, not a variable or function call'
)
_IF_CONDITIONS_ARE_NOT_ALLOWED: Final = (
    'Using "if" conditions inside a generator is not allowed'
)


def analyze(ctx: MethodContext) -> MypyType:
    """
    Used to handle validation and error types in :ref:`do-notation`.

    What it does?

    1. For all types we ensure that only a single container type
       is used in a single do-notation. We don't allow mixing them.
    2. For types with error types (like ``Result``),
       it inferes what possible errors types can we have.
       The result is a ``Union`` of all possible errors.
    3. Ensures that expression passed into ``.do`` method is literal.
    4. Checks that default value is provided
       if generator expression has ``if`` conditions inside.

    """
    default_return = get_proper_type(ctx.default_return_type)
    if not ctx.args or not ctx.args[0]:
        return default_return

    expr = ctx.args[0][0]
    if not isinstance(expr, GeneratorExpr):
        ctx.api.fail(_LITERAL_GENERATOR_EXPR_REQUIRED, expr)
        return default_return
    if not isinstance(ctx.type, CallableType):
        return default_return
    if not isinstance(default_return, Instance):
        return default_return

    return _do_notation(
        expr=expr,
        type_info=ctx.type.type_object(),
        default_return_type=default_return,
        ctx=ctx,
    )


def _do_notation(
    expr: GeneratorExpr,
    type_info: TypeInfo,
    default_return_type: Instance,
    ctx: MethodContext,
) -> MypyType:
    types = []
    for seq in expr.sequences:
        error_type = _try_fetch_error_type(type_info, seq, ctx)
        if error_type is not None:
            types.append(error_type)

    _check_if_conditions(expr, ctx)

    if types:
        return default_return_type.copy_modified(
            args=[
                default_return_type.args[0],
                make_simplified_union(types),
                *default_return_type.args[2:],
            ],
        )
    return default_return_type


def _try_fetch_error_type(
    type_info: TypeInfo,
    seq: Expression,
    ctx: MethodContext,
) -> MypyType | None:
    inst = Instance(
        type_info,
        [
            AnyType(TypeOfAny.implementation_artifact)
            for _ in type_info.type_vars
        ],
    )
    typ = ctx.api.expr_checker.accept(seq)  # type: ignore
    if is_subtype(typ, inst, ignore_type_params=True):
        is_success, error_type = _extract_error_type(typ, type_info)
        if is_success:
            return error_type

    ctx.api.fail(
        _INVALID_DO_NOTATION_SOURCE.format(inst, typ),
        seq,
    )
    return None


def _extract_error_type(
    typ: MypyType,
    type_info: TypeInfo,
) -> tuple[bool, MypyType | None]:
    typ = get_proper_type(typ)
    if isinstance(typ, Instance):
        return True, _decide_error_type(
            map_instance_to_supertype(typ, type_info),
        )

    if isinstance(typ, UnionType):
        types = []
        is_success = True
        for type_item in typ.items:
            is_success, error_type = _extract_error_type(type_item, type_info)
            if error_type is not None:
                types.append(error_type)
        if is_success:
            return True, make_simplified_union(types)
    return False, None


def _decide_error_type(typ: Instance) -> MypyType | None:
    if len(typ.args) < 2:
        return None
    if isinstance(get_proper_type(typ.args[1]), AnyType):
        return None
    return typ.args[1]


def _check_if_conditions(
    expr: GeneratorExpr,
    ctx: MethodContext,
) -> None:
    if any(cond for cond in expr.condlists):
        ctx.api.fail(_IF_CONDITIONS_ARE_NOT_ALLOWED, expr)



from collections.abc import Callable
from typing import TypeVar

from returns.interfaces.bimappable import BiMappableN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')

_UpdatedType1 = TypeVar('_UpdatedType1')
_UpdatedType2 = TypeVar('_UpdatedType2')

_BiMappableKind = TypeVar('_BiMappableKind', bound=BiMappableN)


def bimap(
    on_first: Callable[[_FirstType], _UpdatedType1],
    on_second: Callable[[_SecondType], _UpdatedType2],
) -> Kinded[
    Callable[
        [KindN[_BiMappableKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_BiMappableKind, _UpdatedType1, _UpdatedType2, _ThirdType],
    ]
]:
    """
    Maps container on both: first and second arguments.

    Can be used to synchronize state on both success and failure.

    This is how it should be used:

    .. code:: python

        >>> from returns.io import IOSuccess, IOFailure
        >>> from returns.pointfree import bimap

        >>> def first(argument: int) -> float:
        ...     return argument / 2

        >>> def second(argument: str) -> bool:
        ...     return bool(argument)

        >>> assert bimap(first, second)(IOSuccess(1)) == IOSuccess(0.5)
        >>> assert bimap(first, second)(IOFailure('')) == IOFailure(False)

    Note, that this function works
    for all containers with ``.map`` and ``.alt`` methods.
    See :class:`returns.primitives.interfaces.bimappable.BiMappableN`
    for more info.

    """

    @kinded
    def factory(
        container: KindN[_BiMappableKind, _FirstType, _SecondType, _ThirdType],
    ) -> KindN[_BiMappableKind, _UpdatedType1, _UpdatedType2, _ThirdType]:
        return container.map(on_first).alt(on_second)

    return factory



from abc import ABC
from collections.abc import Callable, Generator, Iterator
from functools import wraps
from inspect import FrameInfo
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar, final, overload

from typing_extensions import ParamSpec

from returns.interfaces.specific import io, ioresult
from returns.primitives.container import BaseContainer, container_equality
from returns.primitives.exceptions import UnwrapFailedError
from returns.primitives.hkt import (
    Kind1,
    Kind2,
    SupportsKind1,
    SupportsKind2,
    dekind,
)
from returns.result import Failure, Result, Success

_ValueType_co = TypeVar('_ValueType_co', covariant=True)
_NewValueType = TypeVar('_NewValueType')

_FuncParams = ParamSpec('_FuncParams')

# Result related:
_ErrorType_co = TypeVar('_ErrorType_co', covariant=True)
_NewErrorType = TypeVar('_NewErrorType')


class IO(  # type: ignore[type-var]
    BaseContainer,
    SupportsKind1['IO', _ValueType_co],
    io.IOLike1[_ValueType_co],
):
    """
    Explicit container for impure function results.

    We also sometimes call it "marker" since once it is marked,
    it cannot be ever unmarked.
    There's no way to directly get its internal value.

    Note that ``IO`` represents a computation that never fails.

    Examples of such computations are:

    - read / write to localStorage
    - get the current time
    - write to the console
    - get a random number

    Use ``IOResult[...]`` for operations that might fail.
    Like DB access or network operations.

    See also:
        - https://dev.to/gcanti/getting-started-with-fp-ts-io-36p6
        - https://gist.github.com/chris-taylor/4745921

    """

    __slots__ = ()

    _inner_value: _ValueType_co

    #: Typesafe equality comparison with other `Result` objects.
    equals = container_equality

    def __init__(self, inner_value: _ValueType_co) -> None:
        """
        Public constructor for this type. Also required for typing.

        .. code:: python

          >>> from returns.io import IO
          >>> assert str(IO(1)) == '<IO: 1>'

        """
        super().__init__(inner_value)

    def map(
        self,
        function: Callable[[_ValueType_co], _NewValueType],
    ) -> 'IO[_NewValueType]':
        """
        Applies function to the inner value.

        Applies 'function' to the contents of the IO instance
        and returns a new IO object containing the result.
        'function' should accept a single "normal" (non-container) argument
        and return a non-container result.

        .. code:: python

          >>> def mappable(string: str) -> str:
          ...      return string + 'b'

          >>> assert IO('a').map(mappable) == IO('ab')

        """
        return IO(function(self._inner_value))

    def apply(
        self,
        container: Kind1['IO', Callable[[_ValueType_co], _NewValueType]],
    ) -> 'IO[_NewValueType]':
        """
        Calls a wrapped function in a container on this container.

        .. code:: python

          >>> from returns.io import IO
          >>> assert IO('a').apply(IO(lambda inner: inner + 'b')) == IO('ab')

        Or more complex example that shows how we can work
        with regular functions and multiple ``IO`` arguments:

        .. code:: python

          >>> from returns.curry import curry

          >>> @curry
          ... def appliable(first: str, second: str) -> str:
          ...      return first + second

          >>> assert IO('b').apply(IO('a').apply(IO(appliable))) == IO('ab')

        """
        return self.map(dekind(container)._inner_value)  # noqa: SLF001

    def bind(
        self,
        function: Callable[[_ValueType_co], Kind1['IO', _NewValueType]],
    ) -> 'IO[_NewValueType]':
        """
        Applies 'function' to the result of a previous calculation.

        'function' should accept a single "normal" (non-container) argument
        and return ``IO`` type object.

        .. code:: python

          >>> def bindable(string: str) -> IO[str]:
          ...      return IO(string + 'b')

          >>> assert IO('a').bind(bindable) == IO('ab')

        """
        return dekind(function(self._inner_value))

    #: Alias for `bind` method. Part of the `IOLikeN` interface.
    bind_io = bind

    def __iter__(self) -> Iterator[_ValueType_co]:
        """API for :ref:`do-notation`."""
        yield self._inner_value

    @classmethod
    def do(
        cls,
        expr: Generator[_NewValueType, None, None],
    ) -> 'IO[_NewValueType]':
        """
        Allows working with unwrapped values of containers in a safe way.

        .. code:: python

          >>> from returns.io import IO
          >>> assert IO.do(
          ...     first + second
          ...     for first in IO(2)
          ...     for second in IO(3)
          ... ) == IO(5)

        See :ref:`do-notation` to learn more.

        """
        return IO(next(expr))

    @classmethod
    def from_value(cls, inner_value: _NewValueType) -> 'IO[_NewValueType]':
        """
        Unit function to construct new ``IO`` values.

        Is the same as regular constructor:

        .. code:: python

          >>> from returns.io import IO
          >>> assert IO(1) == IO.from_value(1)

        Part of the :class:`returns.interfaces.applicative.ApplicativeN`
        interface.
        """
        return IO(inner_value)

    @classmethod
    def from_io(cls, inner_value: 'IO[_NewValueType]') -> 'IO[_NewValueType]':
        """
        Unit function to construct new ``IO`` values from existing ``IO``.

        .. code:: python

          >>> from returns.io import IO
          >>> assert IO(1) == IO.from_io(IO(1))

        Part of the :class:`returns.interfaces.specific.IO.IOLikeN` interface.

        """
        return inner_value

    @classmethod
    def from_ioresult(
        cls,
        inner_value: 'IOResult[_NewValueType, _NewErrorType]',
    ) -> 'IO[Result[_NewValueType, _NewErrorType]]':
        """
        Converts ``IOResult[a, b]`` back to ``IO[Result[a, b]]``.

        Can be really helpful for composition.

        .. code:: python

            >>> from returns.io import IO, IOSuccess
            >>> from returns.result import Success
            >>> assert IO.from_ioresult(IOSuccess(1)) == IO(Success(1))

        Is the reverse of :meth:`returns.io.IOResult.from_typecast`.
        """
        return IO(inner_value._inner_value)  # noqa: SLF001


# Helper functions:


def impure(
    function: Callable[_FuncParams, _NewValueType],
) -> Callable[_FuncParams, IO[_NewValueType]]:
    """
    Decorator to mark function that it returns :class:`~IO` container.

    If you need to mark ``async`` function as impure,
    use :func:`returns.future.future` instead.
    This decorator only works with sync functions. Example:

    .. code:: python

      >>> from returns.io import IO, impure

      >>> @impure
      ... def function(arg: int) -> int:
      ...     return arg + 1  # this action is pure, just an example
      ...

      >>> assert function(1) == IO(2)

    """

    @wraps(function)
    def decorator(
        *args: _FuncParams.args,
        **kwargs: _FuncParams.kwargs,
    ) -> IO[_NewValueType]:
        return IO(function(*args, **kwargs))

    return decorator


# IO and Result:


class IOResult(  # type: ignore[type-var]
    BaseContainer,
    SupportsKind2['IOResult', _ValueType_co, _ErrorType_co],
    ioresult.IOResultBased2[_ValueType_co, _ErrorType_co],
    ABC,
):
    """
    Explicit container for impure function results that might fail.

    .. rubric:: Definition

    This type is similar to :class:`returns.result.Result`.
    This basically a more useful version of ``IO[Result[a, b]]``.
    Use this type for ``IO`` computations that might fail.
    Examples of ``IO`` computations that might fail are:

    - access database
    - access network
    - access filesystem

    Use :class:`~IO` for operations that do ``IO`` but do not fail.

    Note, that even methods like :meth:`~IOResult.unwrap``
    and :meth:`~IOResult.value_or` return values wrapped in ``IO``.

    ``IOResult`` is a complex compound value that consists of:

    - raw value
    - ``Result``
    - ``IO``

    This is why it has so many helper and factory methods:

    - You can construct ``IOResult`` from raw values
      with :func:`~IOSuccess` and :func:`~IOFailure` public type constructors
    - You can construct ``IOResult`` from ``IO`` values
      with :meth:`~IOResult.from_failed_io`
      and :meth:`IOResult.from_io`
    - You can construct ``IOResult`` from ``Result`` values
      with :meth:`~IOResult.from_result`

    We also have a lot of utility methods for better function composition like:

    - :meth:`~IOResult.bind_result` to work
      with functions which return ``Result``
    - :meth:`~IOResult.from_typecast` to work with ``IO[Result[...]]`` values

    See also:
        https://github.com/gcanti/fp-ts/blob/master/docs/modules/IOEither.ts.md

    .. rubric:: Implementation

    This class contains all the methods that can be delegated to ``Result``.
    But, some methods are not implemented which means
    that we have to use special :class:`~_IOSuccess` and :class:`~_IOFailure`
    implementation details to correctly handle these callbacks.

    Do not rely on them! Use public functions and types instead.

    """

    __slots__ = ()

    _inner_value: Result[_ValueType_co, _ErrorType_co]
    __match_args__ = ('_inner_value',)

    #: Typesafe equality comparison with other `IOResult` objects.
    equals = container_equality

    def __init__(
        self, inner_value: Result[_ValueType_co, _ErrorType_co]
    ) -> None:
        """
        Private type constructor.

        Use :func:`~IOSuccess` and :func:`~IOFailure` instead.
        Or :meth:`~IOResult.from_result` factory.
        """
        super().__init__(inner_value)

    def __repr__(self) -> str:
        """
        Custom ``str`` representation for better readability.

        .. code:: python

          >>> from returns.io import IOSuccess, IOFailure
          >>> assert str(IOSuccess(1)) == '<IOResult: <Success: 1>>'
          >>> assert repr(IOSuccess(1)) == '<IOResult: <Success: 1>>'
          >>> str(IOFailure(ValueError('wrong!')))
          '<IOResult: <Failure: wrong!>>'

        """
        return f'<IOResult: {self._inner_value}>'

    @property
    def trace(self) -> list[FrameInfo] | None:
        """Returns a stack trace when :func:`~IOFailure` was called."""
        return self._inner_value.trace

    def swap(self) -> 'IOResult[_ErrorType_co, _ValueType_co]':
        """
        Swaps value and error types.

        So, values become errors and errors become values.
        It is useful when you have to work with errors a lot.
        And since we have a lot of ``.bind_`` related methods
        and only a single ``.lash`` - it is easier to work with values.

        .. code:: python

          >>> from returns.io import IOSuccess, IOFailure
          >>> assert IOSuccess(1).swap() == IOFailure(1)
          >>> assert IOFailure(1).swap() == IOSuccess(1)

        """
        return self.from_result(self._inner_value.swap())

    def map(
        self,
        function: Callable[[_ValueType_co], _NewValueType],
    ) -> 'IOResult[_NewValueType, _ErrorType_co]':
        """
        Composes successful container with a pure function.

        .. code:: python

          >>> from returns.io import IOSuccess
          >>> assert IOSuccess(1).map(lambda num: num + 1) == IOSuccess(2)

        """
        return self.from_result(self._inner_value.map(function))

    def apply(
        self,
        container: Kind2[
            'IOResult',
            Callable[[_ValueType_co], _NewValueType],
            _ErrorType_co,
        ],
    ) -> 'IOResult[_NewValueType, _ErrorType_co]':
        """
        Calls a wrapped function in a container on this container.

        .. code:: python

          >>> from returns.io import IOSuccess, IOFailure

          >>> def appliable(first: str) -> str:
          ...      return first + 'b'

          >>> assert IOSuccess('a').apply(
          ...     IOSuccess(appliable),
          ... ) == IOSuccess('ab')
          >>> assert IOFailure('a').apply(
          ...     IOSuccess(appliable),
          ... ) == IOFailure('a')

          >>> assert IOSuccess('a').apply(IOFailure(1)) == IOFailure(1)
          >>> assert IOFailure('a').apply(IOFailure('b')) == IOFailure('a')

        """
        if isinstance(self, IOFailure):
            return self
        if isinstance(container, IOSuccess):
            return self.from_result(
                self._inner_value.map(
                    container.unwrap()._inner_value,  # noqa: SLF001
                ),
            )
        return container  # type: ignore

    def bind(
        self,
        function: Callable[
            [_ValueType_co],
            Kind2['IOResult', _NewValueType, _ErrorType_co],
        ],
    ) -> 'IOResult[_NewValueType, _ErrorType_co]':
        """
        Composes successful container with a function that returns a container.

        .. code:: python

          >>> from returns.io import IOResult, IOFailure, IOSuccess
          >>> def bindable(string: str) -> IOResult[str, str]:
          ...      if len(string) > 1:
          ...          return IOSuccess(string + 'b')
          ...      return IOFailure(string + 'c')

          >>> assert IOSuccess('aa').bind(bindable) == IOSuccess('aab')
          >>> assert IOSuccess('a').bind(bindable) == IOFailure('ac')
          >>> assert IOFailure('a').bind(bindable) == IOFailure('a')

        """

    #: Alias for `bind_ioresult` method. Part of the `IOResultBasedN` interface.
    bind_ioresult = bind

    def bind_result(
        self,
        function: Callable[
            [_ValueType_co],
            Result[_NewValueType, _ErrorType_co],
        ],
    ) -> 'IOResult[_NewValueType, _ErrorType_co]':
        """
        Composes successful container with a function that returns a container.

        Similar to :meth:`~IOResult.bind`, but works with containers
        that return :class:`returns.result.Result`
        instead of :class:`~IOResult`.

        .. code:: python

          >>> from returns.io import IOFailure, IOSuccess
          >>> from returns.result import Result, Success

          >>> def bindable(string: str) -> Result[str, str]:
          ...      if len(string) > 1:
          ...          return Success(string + 'b')
          ...      return Failure(string + 'c')

          >>> assert IOSuccess('aa').bind_result(bindable) == IOSuccess('aab')
          >>> assert IOSuccess('a').bind_result(bindable) == IOFailure('ac')
          >>> assert IOFailure('a').bind_result(bindable) == IOFailure('a')

        """

    def bind_io(
        self,
        function: Callable[[_ValueType_co], IO[_NewValueType]],
    ) -> 'IOResult[_NewValueType, _ErrorType_co]':
        """
        Composes successful container with a function that returns a container.

        Similar to :meth:`~IOResult.bind`, but works with containers
        that return :class:`returns.io.IO` instead of :class:`~IOResult`.

        .. code:: python

          >>> from returns.io import IO, IOFailure, IOSuccess

          >>> def bindable(string: str) -> IO[str]:
          ...      return IO(string + 'z')

          >>> assert IOSuccess('a').bind_io(bindable) == IOSuccess('az')
          >>> assert IOFailure('a').bind_io(bindable) == IOFailure('a')

        """

    def alt(
        self,
        function: Callable[[_ErrorType_co], _NewErrorType],
    ) -> 'IOResult[_ValueType_co, _NewErrorType]':
        """
        Composes failed container with a pure function to modify failure.

        .. code:: python

          >>> from returns.io import IOFailure
          >>> assert IOFailure(1).alt(float) == IOFailure(1.0)

        """
        return self.from_result(self._inner_value.alt(function))

    def lash(
        self,
        function: Callable[
            [_ErrorType_co],
            Kind2['IOResult', _ValueType_co, _NewErrorType],
        ],
    ) -> 'IOResult[_ValueType_co, _NewErrorType]':
        """
        Composes failed container with a function that returns a container.

        .. code:: python

          >>> from returns.io import IOFailure, IOSuccess, IOResult
          >>> def lashable(state: str) -> IOResult[int, str]:
          ...     if len(state) > 1:
          ...         return IOSuccess(len(state))
          ...     return IOFailure('oops')

          >>> assert IOFailure('a').lash(lashable) == IOFailure('oops')
          >>> assert IOFailure('abc').lash(lashable) == IOSuccess(3)
          >>> assert IOSuccess('a').lash(lashable) == IOSuccess('a')

        """

    def value_or(
        self,
        default_value: _NewValueType,
    ) -> IO[_ValueType_co | _NewValueType]:
        """
        Get value from successful container or default value from failed one.

        .. code:: python

          >>> from returns.io import IO, IOFailure, IOSuccess
          >>> assert IOSuccess(1).value_or(None) == IO(1)
          >>> assert IOFailure(1).value_or(None) == IO(None)

        """
        return IO(self._inner_value.value_or(default_value))

    def unwrap(self) -> IO[_ValueType_co]:
        """
        Get value from successful container or raise exception for failed one.

        .. code:: pycon
          :force:

          >>> from returns.io import IO, IOFailure, IOSuccess
          >>> assert IOSuccess(1).unwrap() == IO(1)

          >>> IOFailure(1).unwrap()
          Traceback (most recent call last):
            ...
          returns.primitives.exceptions.UnwrapFailedError

        """
        return IO(self._inner_value.unwrap())

    def failure(self) -> IO[_ErrorType_co]:
        """
        Get failed value from failed container or raise exception from success.

        .. code:: pycon
          :force:

          >>> from returns.io import IO, IOFailure, IOSuccess
          >>> assert IOFailure(1).failure() == IO(1)

          >>> IOSuccess(1).failure()
          Traceback (most recent call last):
            ...
          returns.primitives.exceptions.UnwrapFailedError

        """
        return IO(self._inner_value.failure())

    def compose_result(
        self,
        function: Callable[
            [Result[_ValueType_co, _ErrorType_co]],
            Kind2['IOResult', _NewValueType, _ErrorType_co],
        ],
    ) -> 'IOResult[_NewValueType, _ErrorType_co]':
        """
        Composes inner ``Result`` with ``IOResult`` returning function.

        Can be useful when you need an access to both states of the result.

        .. code:: python

          >>> from returns.io import IOResult, IOSuccess, IOFailure
          >>> from returns.result import Result

          >>> def count(container: Result[int, int]) -> IOResult[int, int]:
          ...     return IOResult.from_result(
          ...         container.map(lambda x: x + 1).alt(abs),
          ...     )

          >>> assert IOSuccess(1).compose_result(count) == IOSuccess(2)
          >>> assert IOFailure(-1).compose_result(count) == IOFailure(1)

        """
        return dekind(function(self._inner_value))

    def __iter__(self) -> Iterator[_ValueType_co]:
        """API for :ref:`do-notation`."""
        # We also unwrap `IO` here.
        yield self.unwrap()._inner_value  # noqa: SLF001

    @classmethod
    def do(
        cls,
        expr: Generator[_NewValueType, None, None],
    ) -> 'IOResult[_NewValueType, _NewErrorType]':
        """
        Allows working with unwrapped values of containers in a safe way.

        .. code:: python

          >>> from returns.io import IOResult, IOFailure, IOSuccess

          >>> assert IOResult.do(
          ...     first + second
          ...     for first in IOSuccess(2)
          ...     for second in IOSuccess(3)
          ... ) == IOSuccess(5)

          >>> assert IOResult.do(
          ...     first + second
          ...     for first in IOFailure('a')
          ...     for second in IOSuccess(3)
          ... ) == IOFailure('a')

        See :ref:`do-notation` to learn more.
        This feature requires our :ref:`mypy plugin <mypy-plugins>`.

        """
        try:
            return IOResult.from_value(next(expr))
        except UnwrapFailedError as exc:
            return IOResult.from_result(exc.halted_container)  # type: ignore

    @classmethod
    def from_typecast(
        cls,
        inner_value: IO[Result[_NewValueType, _NewErrorType]],
    ) -> 'IOResult[_NewValueType, _NewErrorType]':
        """
        Converts ``IO[Result[_ValueType_co, _ErrorType_co]]`` to ``IOResult``.

        Also prevails the type of ``Result`` to ``IOResult``, so:
        ``IO[Result[_ValueType_co, _ErrorType_co]]`` would become
        ``IOResult[_ValueType_co, _ErrorType_co]``.

        .. code:: python

          >>> from returns.result import Success
          >>> from returns.io import IO, IOResult, IOSuccess
          >>> container = IO(Success(1))
          >>> assert IOResult.from_typecast(container) == IOSuccess(1)

        Can be reverted via :meth:`returns.io.IO.from_ioresult` method.
        """
        return cls.from_result(inner_value._inner_value)  # noqa: SLF001

    @classmethod
    def from_failed_io(
        cls,
        inner_value: IO[_NewErrorType],
    ) -> 'IOResult[Any, _NewErrorType]':
        """
        Creates new ``IOResult`` from "failed" ``IO`` container.

        .. code:: python

          >>> from returns.io import IO, IOResult, IOFailure
          >>> container = IO(1)
          >>> assert IOResult.from_failed_io(container) == IOFailure(1)

        """
        return IOFailure(inner_value._inner_value)  # noqa: SLF001

    @classmethod
    def from_io(
        cls,
        inner_value: IO[_NewValueType],
    ) -> 'IOResult[_NewValueType, Any]':
        """
        Creates new ``IOResult`` from "successful" ``IO`` container.

        .. code:: python

          >>> from returns.io import IO, IOResult, IOSuccess
          >>> container = IO(1)
          >>> assert IOResult.from_io(container) == IOSuccess(1)

        """
        return IOSuccess(inner_value._inner_value)  # noqa: SLF001

    @classmethod
    def from_result(
        cls,
        inner_value: Result[_NewValueType, _NewErrorType],
    ) -> 'IOResult[_NewValueType, _NewErrorType]':
        """
        Creates ``IOResult`` from ``Result`` value.

        .. code:: python

          >>> from returns.io import IOResult, IOSuccess, IOFailure
          >>> from returns.result import Success, Failure

          >>> assert IOResult.from_result(Success(1)) == IOSuccess(1)
          >>> assert IOResult.from_result(Failure(2)) == IOFailure(2)

        """
        if isinstance(inner_value, Success):
            return IOSuccess(inner_value._inner_value)  # noqa: SLF001
        return IOFailure(inner_value._inner_value)  # type: ignore[arg-type]  # noqa: SLF001

    @classmethod
    def from_ioresult(
        cls,
        inner_value: 'IOResult[_NewValueType, _NewErrorType]',
    ) -> 'IOResult[_NewValueType, _NewErrorType]':
        """
        Creates ``IOResult`` from existing ``IOResult`` value.

        .. code:: python

          >>> from returns.io import IOResult, IOSuccess, IOFailure

          >>> assert IOResult.from_ioresult(IOSuccess(1)) == IOSuccess(1)
          >>> assert IOResult.from_ioresult(IOFailure(2)) == IOFailure(2)

        """
        return inner_value

    @classmethod
    def from_value(
        cls,
        inner_value: _NewValueType,
    ) -> 'IOResult[_NewValueType, Any]':
        """
        One more value to create success unit values.

        It is useful as a united way to create a new value from any container.

        .. code:: python

          >>> from returns.io import IOResult, IOSuccess
          >>> assert IOResult.from_value(1) == IOSuccess(1)

        You can use this method or :func:`~IOSuccess`,
        choose the most convenient for you.

        """
        return IOSuccess(inner_value)

    @classmethod
    def from_failure(
        cls,
        inner_value: _NewErrorType,
    ) -> 'IOResult[Any, _NewErrorType]':
        """
        One more value to create failure unit values.

        It is useful as a united way to create a new value from any container.

        .. code:: python

          >>> from returns.io import IOResult, IOFailure
          >>> assert IOResult.from_failure(1) == IOFailure(1)

        You can use this method or :func:`~IOFailure`,
        choose the most convenient for you.

        """
        return IOFailure(inner_value)


@final
class IOFailure(IOResult[Any, _ErrorType_co]):
    """``IOFailure`` representation."""

    __slots__ = ()

    _inner_value: Result[Any, _ErrorType_co]

    def __init__(self, inner_value: _ErrorType_co) -> None:
        """IOFailure constructor."""
        super().__init__(Failure(inner_value))

    if not TYPE_CHECKING:  # noqa: WPS604  # pragma: no branch

        def bind(self, function):
            """Does nothing for ``IOFailure``."""
            return self

        #: Alias for `bind_ioresult` method. Part of the `IOResultBasedN` interface.  # noqa: E501
        bind_ioresult = bind

        def bind_result(self, function):
            """Does nothing for ``IOFailure``."""
            return self

        def bind_io(self, function):
            """Does nothing for ``IOFailure``."""
            return self

        def lash(self, function):
            """Composes this container with a function returning ``IOResult``."""  # noqa: E501
            return function(self._inner_value.failure())


@final
class IOSuccess(IOResult[_ValueType_co, Any]):
    """``IOSuccess`` representation."""

    __slots__ = ()

    _inner_value: Result[_ValueType_co, Any]

    def __init__(self, inner_value: _ValueType_co) -> None:
        """IOSuccess constructor."""
        super().__init__(Success(inner_value))

    if not TYPE_CHECKING:  # noqa: WPS604  # pragma: no branch

        def bind(self, function):
            """Composes this container with a function returning ``IOResult``."""  # noqa: E501
            return function(self._inner_value.unwrap())

        #: Alias for `bind_ioresult` method. Part of the `IOResultBasedN` interface.  # noqa: E501
        bind_ioresult = bind

        def bind_result(self, function):
            """Binds ``Result`` returning function to current container."""
            return self.from_result(function(self._inner_value.unwrap()))

        def bind_io(self, function):
            """Binds ``IO`` returning function to current container."""
            return self.from_io(function(self._inner_value.unwrap()))

        def lash(self, function):
            """Does nothing for ``IOSuccess``."""
            return self


# Aliases:


#: Alias for ``IOResult[_ValueType_co, Exception]``.
IOResultE: TypeAlias = IOResult[_ValueType_co, Exception]


# impure_safe decorator:

_ExceptionType = TypeVar('_ExceptionType', bound=Exception)


@overload
def impure_safe(
    function: Callable[_FuncParams, _NewValueType],
    /,
) -> Callable[_FuncParams, IOResultE[_NewValueType]]: ...


@overload
def impure_safe(
    exceptions: tuple[type[_ExceptionType], ...],
) -> Callable[
    [Callable[_FuncParams, _NewValueType]],
    Callable[_FuncParams, IOResult[_NewValueType, _ExceptionType]],
]: ...


def impure_safe(  # noqa: WPS234
    exceptions: (
        Callable[_FuncParams, _NewValueType] | tuple[type[_ExceptionType], ...]
    ),
) -> (
    Callable[_FuncParams, IOResultE[_NewValueType]]
    | Callable[
        [Callable[_FuncParams, _NewValueType]],
        Callable[_FuncParams, IOResult[_NewValueType, _ExceptionType]],
    ]
):
    """
    Decorator to mark function that it returns :class:`~IOResult` container.

    Should be used with care, since it only catches ``Exception`` subclasses.
    It does not catch ``BaseException`` subclasses.

    If you need to mark ``async`` function as impure,
    use :func:`returns.future.future_safe` instead.
    This decorator only works with sync functions. Example:

    .. code:: python

      >>> from returns.io import IOSuccess, impure_safe

      >>> @impure_safe
      ... def function(arg: int) -> float:
      ...     return 1 / arg
      ...

      >>> assert function(1) == IOSuccess(1.0)
      >>> assert function(0).failure()

    You can also use it with explicit exception types as the first argument:

    .. code:: python

      >>> from returns.io import IOSuccess, IOFailure, impure_safe

      >>> @impure_safe(exceptions=(ZeroDivisionError,))
      ... def might_raise(arg: int) -> float:
      ...     return 1 / arg

      >>> assert might_raise(1) == IOSuccess(1.0)
      >>> assert isinstance(might_raise(0), IOFailure)

    In this case, only exceptions that are explicitly
    listed are going to be caught.

    Similar to :func:`returns.future.future_safe`
    and :func:`returns.result.safe` decorators.
    """

    def factory(
        inner_function: Callable[_FuncParams, _NewValueType],
        inner_exceptions: tuple[type[_ExceptionType], ...],
    ) -> Callable[_FuncParams, IOResult[_NewValueType, _ExceptionType]]:
        @wraps(inner_function)
        def decorator(
            *args: _FuncParams.args,
            **kwargs: _FuncParams.kwargs,
        ) -> IOResult[_NewValueType, _ExceptionType]:
            try:
                return IOSuccess(inner_function(*args, **kwargs))
            except inner_exceptions as exc:
                return IOFailure(exc)

        return decorator

    if isinstance(exceptions, tuple):
        return lambda function: factory(function, exceptions)
    return factory(
        exceptions,
        (Exception,),  # type: ignore[arg-type]
    )



import pytest

from returns.io import IO, IOFailure, IOResult, IOSuccess


def test_io_map():
    """Ensures that IO container supports ``.map()`` method."""
    io: IO[float] = IO(1).map(
        lambda number: number / 2,
    )

    assert io == IO(0.5)


def test_io_bind():
    """Ensures that IO container supports ``.bind()`` method."""
    io: IO[int] = IO('1').bind(
        lambda number: IO(int(number)),
    )

    assert io == IO(1)


def test_io_str():
    """Ensures that IO container supports str cast."""
    assert str(IO([])) == '<IO: []>'


@pytest.mark.parametrize(
    'container',
    [
        IOSuccess(1),
        IOFailure(1),
    ],
)
def test_io_typecast_reverse(container):
    """Ensures that IO can be casted to IOResult and back."""
    assert IO.from_ioresult(container) == IO.from_ioresult(
        IOResult.from_typecast(IO.from_ioresult(container)),
    )



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, TypeVar

from returns.interfaces.specific.ioresult import IOResultLikeN
from returns.primitives.hkt import Kinded, KindN, kinded

if TYPE_CHECKING:
    from returns.io import IOResult  # noqa: WPS433

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_IOResultLikeKind = TypeVar('_IOResultLikeKind', bound=IOResultLikeN)


def bind_ioresult(
    function: Callable[[_FirstType], IOResult[_UpdatedType, _SecondType]],
) -> Kinded[
    Callable[
        [KindN[_IOResultLikeKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_IOResultLikeKind, _UpdatedType, _SecondType, _ThirdType],
    ]
]:
    """
    Composes successful container with a function that returns a container.

    In other words, it modifies the function's
    signature from:
    ``a -> IOResult[b, c]``
    to:
    ``Container[a, c] -> Container[b, c]``

    .. code:: python

      >>> from returns.io import IOResult, IOSuccess
      >>> from returns.context import RequiresContextIOResult
      >>> from returns.pointfree import bind_ioresult

      >>> def returns_ioresult(arg: int) -> IOResult[int, str]:
      ...     return IOSuccess(arg + 1)

      >>> bound = bind_ioresult(returns_ioresult)
      >>> assert bound(IOSuccess(1)) == IOSuccess(2)
      >>> assert bound(
      ...     RequiresContextIOResult.from_value(1),
      ... )(...) == IOSuccess(2)

    """

    @kinded
    def factory(
        container: KindN[
            _IOResultLikeKind,
            _FirstType,
            _SecondType,
            _ThirdType,
        ],
    ) -> KindN[_IOResultLikeKind, _UpdatedType, _SecondType, _ThirdType]:
        return container.bind_ioresult(function)

    return factory



from collections.abc import Sequence
from typing import Final, cast

import anyio  # you would need to `pip install anyio`
import httpx  # you would need to `pip install httpx`
from typing_extensions import TypedDict

from returns.context import RequiresContextFutureResultE
from returns.functions import tap
from returns.future import FutureResult, future_safe
from returns.iterables import Fold
from returns.pipeline import managed
from returns.result import ResultE, safe

_URL: Final = 'https://jsonplaceholder.typicode.com/posts/{0}'


class _Post(TypedDict):
    id: int
    user_id: int
    title: str
    body: str


def _close(
    client: httpx.AsyncClient,
    raw_value: ResultE[Sequence[str]],
) -> FutureResult[None, Exception]:
    return future_safe(client.aclose)()


def _fetch_post(
    post_id: int,
) -> RequiresContextFutureResultE[_Post, httpx.AsyncClient]:
    context: RequiresContextFutureResultE[
        httpx.AsyncClient,
        httpx.AsyncClient,
    ] = RequiresContextFutureResultE.ask()

    return (
        context
        .bind_future_result(
            lambda client: future_safe(client.get)(_URL.format(post_id)),
        )
        .bind_result(
            safe(tap(httpx.Response.raise_for_status)),
        )
        .map(
            lambda response: cast(_Post, response.json()),  # or validate it
        )
    )


def _show_titles(
    number_of_posts: int,
) -> RequiresContextFutureResultE[Sequence[str], httpx.AsyncClient]:
    def factory(post: _Post) -> str:  # noqa: FURB118
        return post['title']

    titles = [
        # Notice how easily we compose async and sync functions:
        _fetch_post(post_id).map(factory)
        # TODO: try `for post_id in (2, 1, 0):` to see how errors work
        for post_id in range(1, number_of_posts + 1)
    ]
    return Fold.collect(titles, RequiresContextFutureResultE.from_value(()))


if __name__ == '__main__':
    # Let's fetch 3 titles of posts one-by-one, but with async client,
    # because we want to highlight `managed` in this example:
    managed_httpx = managed(_show_titles(3), _close)
    future_result = managed_httpx(
        FutureResult.from_value(httpx.AsyncClient(timeout=5)),
    )
    print(anyio.run(future_result.awaitable))  # noqa: WPS421
    # <IOResult: <Success: (
    #    'sunt aut facere repellat provident occaecati ...',
    #    'qui est esse',
    #    'ea molestias quasi exercitationem repellat qui ipsa sit aut',
    # )>>



from abc import ABC
from typing import Any, TypeVar

from typing_extensions import TypedDict

from returns.interfaces.equable import Equable
from returns.primitives.hkt import Kind1
from returns.primitives.types import Immutable

_EqualType = TypeVar('_EqualType', bound=Equable)


class _PickleState(TypedDict):
    """Dict to represent the `BaseContainer` state to be pickled."""

    # TODO: Remove `__slots__` from here when `slotscheck` allow ignore classes
    # by using comments. We don't need the slots here since this class is just
    # a representation of a dictionary and should not be instantiated by any
    # means.
    # See: https://github.com/ariebovenberg/slotscheck/issues/71
    __slots__ = ('container_value',)  # type: ignore

    container_value: Any


class BaseContainer(Immutable, ABC):
    """Utility class to provide all needed magic methods to the context."""

    __slots__ = ('_inner_value',)
    _inner_value: Any

    def __init__(self, inner_value) -> None:
        """
        Wraps the given value in the Container.

        'value' is any arbitrary value of any type including functions.
        """
        object.__setattr__(self, '_inner_value', inner_value)

    def __repr__(self) -> str:
        """Used to display details of object."""
        return '<{}: {}>'.format(
            self.__class__.__qualname__.strip('_'),
            str(self._inner_value),
        )

    def __eq__(self, other: object) -> bool:
        """Used to compare two 'Container' objects."""
        return container_equality(self, other)  # type: ignore

    def __hash__(self) -> int:
        """Used to use this value as a key."""
        return hash(self._inner_value)

    def __getstate__(self) -> _PickleState:
        """That's how this object will be pickled."""
        return {'container_value': self._inner_value}  # type: ignore

    def __setstate__(self, state: _PickleState | Any) -> None:
        """Loading state from pickled data."""
        if isinstance(state, dict) and 'container_value' in state:
            object.__setattr__(
                self,
                '_inner_value',
                state['container_value'],
            )
        else:
            # backward compatibility with 0.19.0 and earlier
            object.__setattr__(self, '_inner_value', state)


def container_equality(
    self: Kind1[_EqualType, Any],
    other: Kind1[_EqualType, Any],
) -> bool:
    """
    Function to compare similar containers.

    Compares both their types and their inner values.
    """
    if type(self) != type(other):  # noqa: WPS516, E721
        return False
    return bool(
        self._inner_value == other._inner_value,  # type: ignore # noqa: SLF001
    )



from returns.maybe import Maybe, Nothing, Some


def test_bind_some():
    """Ensures that bind works correctly."""

    def factory(inner_value: int) -> Maybe[int]:
        return Some(inner_value * 2)

    input_value = 5
    bound = Some(input_value).bind(factory)

    assert bound == factory(input_value)
    assert str(bound) == '<Some: 10>'


def test_bind_optional():
    """Ensures that bind_optional works correctly."""

    def factory(inner_value: int) -> int | None:
        return inner_value or None

    assert Some(1).bind_optional(factory) == Some(1)
    assert Some(0).bind_optional(factory) == Nothing
    assert Nothing.bind_optional(factory) == Nothing



from collections.abc import Callable
from typing import TypeVar

from returns.interfaces.failable import DiverseFailableN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_NewFirstType = TypeVar('_NewFirstType')
_SecondType = TypeVar('_SecondType')
_NewSecondType = TypeVar('_NewSecondType')
_ThirdType = TypeVar('_ThirdType')
_NewThirdType = TypeVar('_NewThirdType')

_DiverseFailableKind = TypeVar('_DiverseFailableKind', bound=DiverseFailableN)


def unify(  # noqa: WPS234
    function: Callable[
        [_FirstType],
        KindN[
            _DiverseFailableKind,
            _NewFirstType,
            _NewSecondType,
            _NewThirdType,
        ],
    ],
) -> Kinded[
    Callable[
        [KindN[_DiverseFailableKind, _FirstType, _SecondType, _ThirdType]],
        KindN[
            _DiverseFailableKind,
            _NewFirstType,
            _SecondType | _NewSecondType,
            _NewThirdType,
        ],
    ]
]:
    """
    Composes successful container with a function that returns a container.

    Similar to :func:`~returns.pointfree.bind` but has different type.
    It returns ``Result[ValueType, Union[OldErrorType, NewErrorType]]``
    instead of ``Result[ValueType, OldErrorType]``.

    So, it can be more useful in some situations.
    Probably with specific exceptions.

    .. code:: python

      >>> from returns.methods import cond
      >>> from returns.pointfree import unify
      >>> from returns.result import Result, Success, Failure

      >>> def bindable(arg: int) -> Result[int, int]:
      ...     return cond(Result, arg % 2 == 0, arg + 1, arg - 1)

      >>> assert unify(bindable)(Success(2)) == Success(3)
      >>> assert unify(bindable)(Success(1)) == Failure(0)
      >>> assert unify(bindable)(Failure(42)) == Failure(42)

    """

    @kinded
    def factory(
        container: KindN[
            _DiverseFailableKind,
            _FirstType,
            _SecondType,
            _ThirdType,
        ],
    ) -> KindN[
        _DiverseFailableKind,
        _NewFirstType,
        _SecondType | _NewSecondType,
        _NewThirdType,
    ]:
        return container.bind(function)  # type: ignore

    return factory



from collections.abc import Callable
from typing import TypeVar

from returns.interfaces.specific.reader import ReaderLike2, ReaderLike3
from returns.primitives.hkt import Kind2, Kind3, Kinded, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_Reader2Kind = TypeVar('_Reader2Kind', bound=ReaderLike2)
_Reader3Kind = TypeVar('_Reader3Kind', bound=ReaderLike3)


def modify_env2(
    function: Callable[[_UpdatedType], _SecondType],
) -> Kinded[
    Callable[
        [Kind2[_Reader2Kind, _FirstType, _SecondType]],
        Kind2[_Reader2Kind, _FirstType, _UpdatedType],
    ]
]:
    """
    Modifies the second type argument of a ``ReaderLike2``.

    In other words, it modifies the function's
    signature from:
    ``a -> b``
    to:
    ``Container[x, a] -> Container[x, b]``

    .. code:: python

      >>> from returns.pointfree import modify_env2
      >>> from returns.context import RequiresContext

      >>> def multiply(arg: int) -> RequiresContext[int, int]:
      ...     return RequiresContext(lambda deps: arg * deps)

      >>> assert modify_env2(int)(multiply(3))('4') == 12

    Note, that this function works with only ``Kind2`` containers
    with ``.modify_env`` method.
    See :class:`returns.primitives.interfaces.specific.reader.ReaderLike2`
    for more info.

    """

    @kinded
    def factory(
        container: Kind2[_Reader2Kind, _FirstType, _SecondType],
    ) -> Kind2[_Reader2Kind, _FirstType, _UpdatedType]:
        return container.modify_env(function)

    return factory


def modify_env3(
    function: Callable[[_UpdatedType], _ThirdType],
) -> Kinded[
    Callable[
        [Kind3[_Reader3Kind, _FirstType, _SecondType, _ThirdType]],
        Kind3[_Reader3Kind, _FirstType, _SecondType, _UpdatedType],
    ]
]:
    """
    Modifies the third type argument of a ``ReaderLike3``.

    In other words, it modifies the function's
    signature from: ``a -> b``
    to: ``Container[x, a] -> Container[x, b]``

    .. code:: python

      >>> from returns.pointfree import modify_env
      >>> from returns.context import RequiresContextResultE
      >>> from returns.result import Success, safe

      >>> def divide(arg: int) -> RequiresContextResultE[float, int]:
      ...     return RequiresContextResultE(safe(lambda deps: arg / deps))

      >>> assert modify_env(int)(divide(3))('2') == Success(1.5)
      >>> assert modify_env(int)(divide(3))('0').failure()

    Note, that this function works with only ``Kind3`` containers
    with ``.modify_env`` method.
    See :class:`returns.primitives.interfaces.specific.reader.ReaderLike3`
    for more info.

    """

    @kinded
    def factory(
        container: Kind3[_Reader3Kind, _FirstType, _SecondType, _ThirdType],
    ) -> Kind3[_Reader3Kind, _FirstType, _SecondType, _UpdatedType]:
        return container.modify_env(function)

    return factory


#: Useful alias for :func:`~modify_env3`.
modify_env = modify_env3



"""
Represents the base interfaces for types that do fear-some async operations.

This type means that ``FutureResult`` can (and will!) fail with exceptions.

Use this type to mark that this specific async opetaion can fail.
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Never, Self, TypeVar

from returns.interfaces.specific import future, ioresult
from returns.primitives.hkt import KindN

if TYPE_CHECKING:
    from returns.future import Future, FutureResult  # noqa: WPS433

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_ValueType = TypeVar('_ValueType')
_ErrorType = TypeVar('_ErrorType')

_FutureResultLikeType = TypeVar(
    '_FutureResultLikeType',
    bound='FutureResultLikeN',
)


class FutureResultLikeN(
    future.FutureLikeN[_FirstType, _SecondType, _ThirdType],
    ioresult.IOResultLikeN[_FirstType, _SecondType, _ThirdType],
):
    """
    Base type for ones that does look like ``FutureResult``.

    But at the time this is not a real ``Future`` and cannot be awaited.
    It is also cannot be unwrapped, because it is not a real ``IOResult``.
    """

    __slots__ = ()

    @abstractmethod
    def bind_future_result(
        self: _FutureResultLikeType,
        function: Callable[
            [_FirstType],
            FutureResult[_UpdatedType, _SecondType],
        ],
    ) -> KindN[_FutureResultLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Allows to bind ``FutureResult`` functions over a container."""

    @abstractmethod
    def bind_async_future_result(
        self: _FutureResultLikeType,
        function: Callable[
            [_FirstType],
            Awaitable[FutureResult[_UpdatedType, _SecondType]],
        ],
    ) -> KindN[_FutureResultLikeType, _UpdatedType, _SecondType, _ThirdType]:
        """Allows to bind async ``FutureResult`` functions over container."""

    @classmethod
    @abstractmethod
    def from_failed_future(
        cls: type[_FutureResultLikeType],
        inner_value: Future[_ErrorType],
    ) -> KindN[_FutureResultLikeType, _FirstType, _ErrorType, _ThirdType]:
        """Creates new container from a failed ``Future``."""

    @classmethod
    def from_future_result(
        cls,
        inner_value: FutureResult[_ValueType, _ErrorType],
    ) -> KindN[Self, _ValueType, _ErrorType, _ThirdType]:
        """Creates container from ``FutureResult`` instance."""


#: Type alias for kinds with two type arguments.
FutureResultLike2 = FutureResultLikeN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
FutureResultLike3 = FutureResultLikeN[_FirstType, _SecondType, _ThirdType]


class FutureResultBasedN(
    future.FutureBasedN[_FirstType, _SecondType, _ThirdType],
    FutureResultLikeN[_FirstType, _SecondType, _ThirdType],
):
    """
    Base type for real ``FutureResult`` objects.

    They can be awaited.
    Still cannot be unwrapped.
    """

    __slots__ = ()


#: Type alias for kinds with two type arguments.
FutureResultBased2 = FutureResultBasedN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
FutureResultBased3 = FutureResultBasedN[_FirstType, _SecondType, _ThirdType]



from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, TypeVar

from returns.io import IO
from returns.primitives.hkt import Kind1, dekind

if TYPE_CHECKING:
    from returns.future import Future

_ValueType_co = TypeVar('_ValueType_co', covariant=True)
_NewValueType = TypeVar('_NewValueType')


async def async_map(
    function: Callable[[_ValueType_co], _NewValueType],
    inner_value: Awaitable[_ValueType_co],
) -> _NewValueType:
    """Async maps a function over a value."""
    return function(await inner_value)


async def async_apply(
    container: 'Future[Callable[[_ValueType_co], _NewValueType]]',
    inner_value: Awaitable[_ValueType_co],
) -> _NewValueType:
    """Async applies a container with function over a value."""
    return (await container)._inner_value(await inner_value)  # noqa: SLF001


async def async_bind(
    function: Callable[[_ValueType_co], Kind1['Future', _NewValueType]],
    inner_value: Awaitable[_ValueType_co],
) -> _NewValueType:
    """Async binds a container over a value."""
    return (await dekind(function(await inner_value)))._inner_value  # noqa: SLF001


async def async_bind_awaitable(
    function: Callable[[_ValueType_co], Awaitable[_NewValueType]],
    inner_value: Awaitable[_ValueType_co],
) -> _NewValueType:
    """Async binds a coroutine over a value."""
    return await function(await inner_value)


async def async_bind_async(
    function: Callable[
        [_ValueType_co],
        Awaitable[Kind1['Future', _NewValueType]],
    ],
    inner_value: Awaitable[_ValueType_co],
) -> _NewValueType:
    """Async binds a coroutine with container over a value."""
    inner_io = dekind(await function(await inner_value))._inner_value  # noqa: SLF001
    return await inner_io


async def async_bind_io(
    function: Callable[[_ValueType_co], IO[_NewValueType]],
    inner_value: Awaitable[_ValueType_co],
) -> _NewValueType:
    """Async binds a container over a value."""
    return function(await inner_value)._inner_value  # noqa: SLF001



import pytest

from returns.maybe import Nothing, Some
from returns.primitives.exceptions import UnwrapFailedError


def test_unwrap_success():
    """Ensures that unwrap works for Some container."""
    assert Some(5).unwrap() == 5


def test_unwrap_failure():
    """Ensures that unwrap works for Nothing container."""
    with pytest.raises(UnwrapFailedError):
        assert Nothing.unwrap()



from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, ClassVar, TypeAlias, TypeVar, final

from returns._internal.futures import _reader_future_result
from returns.context import NoDeps
from returns.future import Future, FutureResult
from returns.interfaces.specific import future_result, reader_future_result
from returns.io import IO, IOResult
from returns.primitives.container import BaseContainer
from returns.primitives.hkt import Kind3, SupportsKind3, dekind
from returns.result import Result

if TYPE_CHECKING:
    from returns.context.requires_context import RequiresContext
    from returns.context.requires_context_ioresult import (
        ReaderIOResult,
        RequiresContextIOResult,
    )
    from returns.context.requires_context_result import RequiresContextResult

# Context:
_EnvType_contra = TypeVar('_EnvType_contra', contravariant=True)
_NewEnvType = TypeVar('_NewEnvType')

# Result:
_ValueType_co = TypeVar('_ValueType_co', covariant=True)
_NewValueType = TypeVar('_NewValueType')
_ErrorType_co = TypeVar('_ErrorType_co', covariant=True)
_NewErrorType = TypeVar('_NewErrorType')

# Helpers:
_FirstType = TypeVar('_FirstType')


@final
class RequiresContextFutureResult(  # type: ignore[type-var]
    BaseContainer,
    SupportsKind3[
        'RequiresContextFutureResult',
        _ValueType_co,
        _ErrorType_co,
        _EnvType_contra,
    ],
    reader_future_result.ReaderFutureResultBasedN[
        _ValueType_co,
        _ErrorType_co,
        _EnvType_contra,
    ],
    future_result.FutureResultLike3[
        _ValueType_co, _ErrorType_co, _EnvType_contra
    ],
):
    """
    The ``RequiresContextFutureResult`` combinator.

    This probably the main type people are going to use in ``async`` programs.

    See :class:`returns.context.requires_context.RequiresContext`,
    :class:`returns.context.requires_context_result.RequiresContextResult`,
    and
    :class:`returns.context.requires_context_result.RequiresContextIOResult`
    for more docs.

    This is just a handy wrapper around
    ``RequiresContext[FutureResult[a, b], env]``
    which represents a context-dependent impure async operation that might fail.

    So, this is a thin wrapper, without any changes in logic.
    Why do we need this wrapper? That's just for better usability!

    This way ``RequiresContextIOResult`` allows to simply work with:

    - raw values and pure functions
    - ``RequiresContext`` values and pure functions returning it
    - ``RequiresContextResult`` values and pure functions returning it
    - ``RequiresContextIOResult`` values and pure functions returning it
    - ``Result`` and pure functions returning it
    - ``IOResult`` and functions returning it
    - ``FutureResult`` and functions returning it
    - other ``RequiresContextFutureResult`` related functions and values

    This is a complex type for complex tasks!

    .. rubric:: Important implementation details

    Due it is meaning, ``RequiresContextFutureResult``
    cannot have ``Success`` and ``Failure`` subclasses.

    We only have just one type. That's by design.

    Different converters are also not supported for this type.
    Use converters inside the ``RequiresContext`` context, not outside.

    See also:
        - https://dev.to/gcanti/getting-started-with-fp-ts-reader-1ie5
        - https://en.wikipedia.org/wiki/Lazy_evaluation
        - https://bit.ly/2R8l4WK
        - https://bit.ly/2RwP4fp

    """

    __slots__ = ()

    #: Inner value of `RequiresContext`
    #: is just a function that returns `FutureResult`.
    #: This field has an extra 'RequiresContext' just because `mypy` needs it.
    _inner_value: Callable[
        [_EnvType_contra], FutureResult[_ValueType_co, _ErrorType_co]
    ]

    #: A convenient placeholder to call methods created by `.from_value()`.
    no_args: ClassVar[NoDeps] = object()

    def __init__(
        self,
        inner_value: Callable[
            [_EnvType_contra], FutureResult[_ValueType_co, _ErrorType_co]
        ],
    ) -> None:
        """
        Public constructor for this type. Also required for typing.

        Only allows functions of kind ``* -> *``
        and returning :class:`returns.result.Result` instances.

        .. code:: python

          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import FutureResult

          >>> instance = RequiresContextFutureResult(
          ...    lambda deps: FutureResult.from_value(1),
          ... )
          >>> str(instance)
          '<RequiresContextFutureResult: <function <lambda> at ...>>'

        """
        super().__init__(inner_value)

    def __call__(
        self, deps: _EnvType_contra
    ) -> FutureResult[_ValueType_co, _ErrorType_co]:
        """
        Evaluates the wrapped function.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import FutureResult
          >>> from returns.io import IOSuccess

          >>> def first(lg: bool) -> RequiresContextFutureResult[int, str, int]:
          ...     # `deps` has `int` type here:
          ...     return RequiresContextFutureResult(
          ...         lambda deps: FutureResult.from_value(
          ...             deps if lg else -deps,
          ...         ),
          ...     )

          >>> instance = first(False)
          >>> assert anyio.run(instance(3).awaitable) == IOSuccess(-3)

          >>> instance = first(True)
          >>> assert anyio.run(instance(3).awaitable) == IOSuccess(3)

        In other things, it is a regular Python magic method.

        """
        return self._inner_value(deps)

    def swap(
        self,
    ) -> RequiresContextFutureResult[
        _ErrorType_co, _ValueType_co, _EnvType_contra
    ]:
        """
        Swaps value and error types.

        So, values become errors and errors become values.
        It is useful when you have to work with errors a lot.
        And since we have a lot of ``.bind_`` related methods
        and only a single ``.lash`` - it is easier to work with values.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> success = RequiresContextFutureResult.from_value(1)
          >>> failure = RequiresContextFutureResult.from_failure(1)

          >>> assert anyio.run(success.swap(), ...) == IOFailure(1)
          >>> assert anyio.run(failure.swap(), ...) == IOSuccess(1)

        """
        return RequiresContextFutureResult(lambda deps: self(deps).swap())

    def map(
        self,
        function: Callable[[_ValueType_co], _NewValueType],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Composes successful container with a pure function.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert anyio.run(RequiresContextFutureResult.from_value(1).map(
          ...     lambda x: x + 1,
          ... )(...).awaitable) == IOSuccess(2)

          >>> assert anyio.run(RequiresContextFutureResult.from_failure(1).map(
          ...     lambda x: x + 1,
          ... )(...).awaitable) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).map(function),
        )

    def apply(
        self,
        container: Kind3[
            RequiresContextFutureResult,
            Callable[[_ValueType_co], _NewValueType],
            _ErrorType_co,
            _EnvType_contra,
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Calls a wrapped function in a container on this container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> def transform(arg: str) -> str:
          ...     return arg + 'b'

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_value('a').apply(
          ...        RequiresContextFutureResult.from_value(transform),
          ...    ),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOSuccess('ab')

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_failure('a').apply(
          ...        RequiresContextFutureResult.from_value(transform),
          ...    ),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOFailure('a')

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).apply(dekind(container)(deps)),
        )

    def bind(
        self,
        function: Callable[
            [_ValueType_co],
            Kind3[
                RequiresContextFutureResult,
                _NewValueType,
                _ErrorType_co,
                _EnvType_contra,
            ],
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Composes this container with a function returning the same type.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import FutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> def function(
          ...     number: int,
          ... ) -> RequiresContextFutureResult[str, int, int]:
          ...     # `deps` has `int` type here:
          ...     return RequiresContextFutureResult(
          ...         lambda deps: FutureResult.from_value(str(number + deps)),
          ...     )

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(2).bind(function),
          ...     3,
          ... ) == IOSuccess('5')
          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failure(2).bind(function),
          ...     3,
          ... ) == IOFailure(2)

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind(
                lambda inner: dekind(  # type: ignore[misc]
                    function(inner),
                )(deps),
            ),
        )

    #: Alias for `bind_context_future_result` method,
    #: it is the same as `bind` here.
    bind_context_future_result = bind

    def bind_async(
        self,
        function: Callable[
            [_ValueType_co],
            Awaitable[
                Kind3[
                    RequiresContextFutureResult,
                    _NewValueType,
                    _ErrorType_co,
                    _EnvType_contra,
                ],
            ],
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Composes this container with a async function returning the same type.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> async def function(
          ...     number: int,
          ... ) -> RequiresContextFutureResult[str, int, int]:
          ...     return RequiresContextFutureResult.from_value(number + 1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(1).bind_async(
          ...        function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(2)
          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failure(1).bind_async(
          ...        function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda deps: FutureResult(
                _reader_future_result.async_bind_async(
                    function,
                    self,
                    deps,
                )
            ),
        )

    #: Alias for `bind_async_context_future_result` method,
    #: it is the same as `bind_async` here.
    bind_async_context_future_result = bind_async

    def bind_awaitable(
        self,
        function: Callable[[_ValueType_co], Awaitable[_NewValueType]],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Allows to compose a container and a regular ``async`` function.

        This function should return plain, non-container value.
        See :meth:`~RequiresContextFutureResult.bind_async`
        to bind ``async`` function that returns a container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> async def coroutine(x: int) -> int:
          ...    return x + 1

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(1).bind_awaitable(
          ...         coroutine,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(2)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failure(1).bind_awaitable(
          ...         coroutine,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind_awaitable(function),
        )

    def bind_result(
        self,
        function: Callable[
            [_ValueType_co], Result[_NewValueType, _ErrorType_co]
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Binds ``Result`` returning function to the current container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.result import Success, Result
          >>> from returns.io import IOSuccess, IOFailure

          >>> def function(num: int) -> Result[int, str]:
          ...     return Success(num + 1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(1).bind_result(
          ...         function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(2)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failure(':(').bind_result(
          ...         function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(':(')

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind_result(function),
        )

    def bind_context(
        self,
        function: Callable[
            [_ValueType_co],
            RequiresContext[_NewValueType, _EnvType_contra],
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Binds ``RequiresContext`` returning function to current container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContext
          >>> from returns.io import IOSuccess, IOFailure

          >>> def function(arg: int) -> RequiresContext[int, str]:
          ...     return RequiresContext(lambda deps: len(deps) + arg)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(2).bind_context(
          ...         function,
          ...     ),
          ...     'abc',
          ... ) == IOSuccess(5)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failure(0).bind_context(
          ...         function,
          ...     ),
          ...     'abc',
          ... ) == IOFailure(0)

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).map(
                lambda inner: function(inner)(deps),  # type: ignore[misc]
            ),
        )

    def bind_context_result(
        self,
        function: Callable[
            [_ValueType_co],
            RequiresContextResult[
                _NewValueType, _ErrorType_co, _EnvType_contra
            ],
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Binds ``RequiresContextResult`` returning function to the current one.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextResult
          >>> from returns.io import IOSuccess, IOFailure
          >>> from returns.result import Success

          >>> def function(arg: int) -> RequiresContextResult[int, int, str]:
          ...     return RequiresContextResult(
          ...         lambda deps: Success(len(deps) + arg),
          ...     )

          >>> instance = RequiresContextFutureResult.from_value(
          ...    2,
          ... ).bind_context_result(
          ...     function,
          ... )('abc')
          >>> assert anyio.run(instance.awaitable) == IOSuccess(5)

          >>> instance = RequiresContextFutureResult.from_failure(
          ...    2,
          ... ).bind_context_result(
          ...     function,
          ... )('abc')
          >>> assert anyio.run(instance.awaitable) == IOFailure(2)

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind_result(
                lambda inner: function(inner)(deps),  # type: ignore[misc]
            ),
        )

    def bind_context_ioresult(
        self,
        function: Callable[
            [_ValueType_co],
            RequiresContextIOResult[
                _NewValueType, _ErrorType_co, _EnvType_contra
            ],
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Binds ``RequiresContextIOResult`` returning function to the current one.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> def function(arg: int) -> RequiresContextIOResult[int, int, str]:
          ...     return RequiresContextIOResult(
          ...         lambda deps: IOSuccess(len(deps) + arg),
          ...     )

          >>> instance = RequiresContextFutureResult.from_value(
          ...    2,
          ... ).bind_context_ioresult(
          ...     function,
          ... )('abc')
          >>> assert anyio.run(instance.awaitable) == IOSuccess(5)

          >>> instance = RequiresContextFutureResult.from_failure(
          ...    2,
          ... ).bind_context_ioresult(
          ...     function,
          ... )('abc')
          >>> assert anyio.run(instance.awaitable) == IOFailure(2)

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind_ioresult(
                lambda inner: function(inner)(deps),  # type: ignore[misc]
            ),
        )

    def bind_io(
        self,
        function: Callable[[_ValueType_co], IO[_NewValueType]],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Binds ``IO`` returning function to the current container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IO, IOSuccess, IOFailure

          >>> def do_io(number: int) -> IO[str]:
          ...     return IO(str(number))  # not IO operation actually

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(1).bind_io(do_io),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess('1')

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failure(1).bind_io(do_io),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind_io(function),
        )

    def bind_ioresult(
        self,
        function: Callable[
            [_ValueType_co], IOResult[_NewValueType, _ErrorType_co]
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Binds ``IOResult`` returning function to the current container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOResult, IOSuccess, IOFailure

          >>> def function(num: int) -> IOResult[int, str]:
          ...     return IOSuccess(num + 1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(1).bind_ioresult(
          ...         function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(2)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failure(':(').bind_ioresult(
          ...         function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(':(')

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind_ioresult(function),
        )

    def bind_future(
        self,
        function: Callable[[_ValueType_co], Future[_NewValueType]],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Binds ``Future`` returning function to the current container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import Future
          >>> from returns.io import IOSuccess, IOFailure

          >>> def function(num: int) -> Future[int]:
          ...     return Future.from_value(num + 1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(1).bind_future(
          ...         function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(2)

          >>> failed = RequiresContextFutureResult.from_failure(':(')
          >>> assert anyio.run(
          ...     failed.bind_future(function),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(':(')

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind_future(function),
        )

    def bind_future_result(
        self,
        function: Callable[
            [_ValueType_co],
            FutureResult[_NewValueType, _ErrorType_co],
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Binds ``FutureResult`` returning function to the current container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import FutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> def function(num: int) -> FutureResult[int, str]:
          ...     return FutureResult.from_value(num + 1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(1).bind_future_result(
          ...         function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(2)

          >>> failed = RequiresContextFutureResult.from_failure(':(')
          >>> assert anyio.run(
          ...     failed.bind_future_result(function),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(':(')

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind(function),
        )

    def bind_async_future(
        self,
        function: Callable[[_ValueType_co], Awaitable[Future[_NewValueType]]],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Binds ``Future`` returning async function to the current container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import Future
          >>> from returns.io import IOSuccess, IOFailure

          >>> async def function(num: int) -> Future[int]:
          ...     return Future.from_value(num + 1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(1).bind_async_future(
          ...         function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(2)

          >>> failed = RequiresContextFutureResult.from_failure(':(')
          >>> assert anyio.run(
          ...     failed.bind_async_future(function),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(':(')

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind_async_future(function),
        )

    def bind_async_future_result(
        self,
        function: Callable[
            [_ValueType_co],
            Awaitable[FutureResult[_NewValueType, _ErrorType_co]],
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Bind ``FutureResult`` returning async function to the current container.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import FutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> async def function(num: int) -> FutureResult[int, str]:
          ...     return FutureResult.from_value(num + 1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(
          ...         1,
          ...     ).bind_async_future_result(
          ...         function,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(2)

          >>> failed = RequiresContextFutureResult.from_failure(':(')
          >>> assert anyio.run(
          ...     failed.bind_async_future_result(function),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(':(')

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).bind_async(function),
        )

    def alt(
        self,
        function: Callable[[_ErrorType_co], _NewErrorType],
    ) -> RequiresContextFutureResult[
        _ValueType_co, _NewErrorType, _EnvType_contra
    ]:
        """
        Composes failed container with a pure function.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value(1).alt(
          ...        lambda x: x + 1,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failure(1).alt(
          ...        lambda x: x + 1,
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(2)

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).alt(function),
        )

    def lash(
        self,
        function: Callable[
            [_ErrorType_co],
            Kind3[
                RequiresContextFutureResult,
                _ValueType_co,
                _NewErrorType,
                _EnvType_contra,
            ],
        ],
    ) -> RequiresContextFutureResult[
        _ValueType_co, _NewErrorType, _EnvType_contra
    ]:
        """
        Composes this container with a function returning the same type.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import FutureResult
          >>> from returns.io import IOSuccess

          >>> def lashable(
          ...     arg: str,
          ... ) -> RequiresContextFutureResult[str, str, str]:
          ...      return RequiresContextFutureResult(
          ...          lambda deps: FutureResult.from_value(
          ...              deps + arg,
          ...          ),
          ...      )

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_value('a').lash(lashable),
          ...     'c',
          ... ) == IOSuccess('a')

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failure('aa').lash(
          ...         lashable,
          ...     ),
          ...     'b',
          ... ) == IOSuccess('baa')

        """
        return RequiresContextFutureResult(
            lambda deps: self(deps).lash(
                lambda inner: function(inner)(deps),  # type: ignore
            ),
        )

    def compose_result(
        self,
        function: Callable[
            [Result[_ValueType_co, _ErrorType_co]],
            Kind3[
                RequiresContextFutureResult,
                _NewValueType,
                _ErrorType_co,
                _EnvType_contra,
            ],
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _ErrorType_co, _EnvType_contra
    ]:
        """
        Composes inner ``Result`` with ``ReaderFutureResult`` returning func.

        Can be useful when you need an access to both states of the result.

        .. code:: python

          >>> import anyio
          >>> from returns.context import ReaderFutureResult, NoDeps
          >>> from returns.io import IOSuccess, IOFailure
          >>> from returns.result import Result

          >>> def count(
          ...    container: Result[int, int],
          ... ) -> ReaderFutureResult[int, int, NoDeps]:
          ...     return ReaderFutureResult.from_result(
          ...         container.map(lambda x: x + 1).alt(abs),
          ...     )

          >>> success = ReaderFutureResult.from_value(1)
          >>> failure = ReaderFutureResult.from_failure(-1)

          >>> assert anyio.run(
          ...     success.compose_result(count), ReaderFutureResult.no_args,
          ... ) == IOSuccess(2)
          >>> assert anyio.run(
          ...     failure.compose_result(count), ReaderFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda deps: FutureResult(
                _reader_future_result.async_compose_result(
                    function,
                    self,
                    deps,
                ),
            ),
        )

    def modify_env(
        self,
        function: Callable[[_NewEnvType], _EnvType_contra],
    ) -> RequiresContextFutureResult[_ValueType_co, _ErrorType_co, _NewEnvType]:
        """
        Allows to modify the environment type.

        .. code:: python

          >>> import anyio
          >>> from returns.future import future_safe, asyncify
          >>> from returns.context import RequiresContextFutureResultE
          >>> from returns.io import IOSuccess

          >>> def div(arg: int) -> RequiresContextFutureResultE[float, int]:
          ...     return RequiresContextFutureResultE(
          ...         future_safe(asyncify(lambda deps: arg / deps)),
          ...     )

          >>> assert anyio.run(div(3).modify_env(int), '2') == IOSuccess(1.5)
          >>> assert anyio.run(div(3).modify_env(int), '0').failure()

        """
        return RequiresContextFutureResult(lambda deps: self(function(deps)))

    @classmethod
    def ask(
        cls,
    ) -> RequiresContextFutureResult[
        _EnvType_contra, _ErrorType_co, _EnvType_contra
    ]:
        """
        Is used to get the current dependencies inside the call stack.

        Similar to
        :meth:`returns.context.requires_context.RequiresContext.ask`,
        but returns ``FutureResult`` instead of a regular value.

        Please, refer to the docs there to learn how to use it.

        One important note that is worth duplicating here:
        you might need to provide type annotations explicitly,
        so ``mypy`` will know about it statically.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResultE
          >>> from returns.io import IOSuccess

          >>> assert anyio.run(
          ...     RequiresContextFutureResultE[int, int].ask().map(str),
          ...     1,
          ... ) == IOSuccess('1')

        """
        return RequiresContextFutureResult(FutureResult.from_value)

    @classmethod
    def from_result(
        cls,
        inner_value: Result[_NewValueType, _NewErrorType],
    ) -> RequiresContextFutureResult[_NewValueType, _NewErrorType, NoDeps]:
        """
        Creates new container with ``Result`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.result import Success, Failure
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_result(Success(1)),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_result(Failure(1)),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda _: FutureResult.from_result(inner_value),
        )

    @classmethod
    def from_io(
        cls,
        inner_value: IO[_NewValueType],
    ) -> RequiresContextFutureResult[_NewValueType, Any, NoDeps]:
        """
        Creates new container from successful ``IO`` value.

        .. code:: python

          >>> import anyio
          >>> from returns.io import IO, IOSuccess
          >>> from returns.context import RequiresContextFutureResult

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_io(IO(1)),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

        """
        return RequiresContextFutureResult(
            lambda deps: FutureResult.from_io(inner_value),
        )

    @classmethod
    def from_failed_io(
        cls,
        inner_value: IO[_NewErrorType],
    ) -> RequiresContextFutureResult[Any, _NewErrorType, NoDeps]:
        """
        Creates a new container from failed ``IO`` value.

        .. code:: python

          >>> import anyio
          >>> from returns.io import IO, IOFailure
          >>> from returns.context import RequiresContextFutureResult

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failed_io(IO(1)),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda deps: FutureResult.from_failed_io(inner_value),
        )

    @classmethod
    def from_ioresult(
        cls,
        inner_value: IOResult[_NewValueType, _NewErrorType],
    ) -> RequiresContextFutureResult[_NewValueType, _NewErrorType, NoDeps]:
        """
        Creates new container with ``IOResult`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_ioresult(IOSuccess(1)),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_ioresult(IOFailure(1)),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda _: FutureResult.from_ioresult(inner_value),
        )

    @classmethod
    def from_future(
        cls,
        inner_value: Future[_NewValueType],
    ) -> RequiresContextFutureResult[_NewValueType, Any, NoDeps]:
        """
        Creates new container with successful ``Future`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import Future
          >>> from returns.io import IOSuccess

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_future(Future.from_value(1)),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

        """
        return RequiresContextFutureResult(
            lambda _: FutureResult.from_future(inner_value),
        )

    @classmethod
    def from_failed_future(
        cls,
        inner_value: Future[_NewErrorType],
    ) -> RequiresContextFutureResult[Any, _NewErrorType, NoDeps]:
        """
        Creates new container with failed ``Future`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import Future
          >>> from returns.io import IOFailure

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_failed_future(
          ...        Future.from_value(1),
          ...    ),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda _: FutureResult.from_failed_future(inner_value),
        )

    @classmethod
    def from_future_result_context(
        cls,
        inner_value: ReaderFutureResult[
            _NewValueType, _NewErrorType, _NewEnvType
        ],
    ) -> ReaderFutureResult[_NewValueType, _NewErrorType, _NewEnvType]:
        """
        Creates new container with ``ReaderFutureResult`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_future_result_context(
          ...        RequiresContextFutureResult.from_value(1),
          ...    ),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_future_result_context(
          ...        RequiresContextFutureResult.from_failure(1),
          ...    ),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return inner_value

    @classmethod
    def from_future_result(
        cls,
        inner_value: FutureResult[_NewValueType, _NewErrorType],
    ) -> RequiresContextFutureResult[_NewValueType, _NewErrorType, NoDeps]:
        """
        Creates new container with ``FutureResult`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.future import FutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_future_result(
          ...        FutureResult.from_value(1),
          ...    ),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

          >>> assert anyio.run(
          ...    RequiresContextFutureResult.from_future_result(
          ...        FutureResult.from_failure(1),
          ...    ),
          ...    RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(lambda _: inner_value)

    @classmethod
    def from_typecast(
        cls,
        inner_value: RequiresContext[
            FutureResult[_NewValueType, _NewErrorType],
            _EnvType_contra,
        ],
    ) -> RequiresContextFutureResult[
        _NewValueType, _NewErrorType, _EnvType_contra
    ]:
        """
        You might end up with ``RequiresContext[FutureResult]`` as a value.

        This method is designed to turn it into ``RequiresContextFutureResult``.
        It will save all the typing information.

        It is just more useful!

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContext
          >>> from returns.future import FutureResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_typecast(
          ...         RequiresContext.from_value(FutureResult.from_value(1)),
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_typecast(
          ...         RequiresContext.from_value(FutureResult.from_failure(1)),
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(inner_value)

    @classmethod
    def from_context(
        cls,
        inner_value: RequiresContext[_NewValueType, _NewEnvType],
    ) -> RequiresContextFutureResult[_NewValueType, Any, _NewEnvType]:
        """
        Creates new container from ``RequiresContext`` as a success unit.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContext
          >>> from returns.io import IOSuccess

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_context(
          ...         RequiresContext.from_value(1),
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

        """
        return RequiresContextFutureResult(
            lambda deps: FutureResult.from_value(inner_value(deps)),
        )

    @classmethod
    def from_failed_context(
        cls,
        inner_value: RequiresContext[_NewValueType, _NewEnvType],
    ) -> RequiresContextFutureResult[Any, _NewValueType, _NewEnvType]:
        """
        Creates new container from ``RequiresContext`` as a failure unit.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContext
          >>> from returns.io import IOFailure

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_failed_context(
          ...         RequiresContext.from_value(1),
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda deps: FutureResult.from_failure(inner_value(deps)),
        )

    @classmethod
    def from_result_context(
        cls,
        inner_value: RequiresContextResult[
            _NewValueType,
            _NewErrorType,
            _NewEnvType,
        ],
    ) -> ReaderFutureResult[_NewValueType, _NewErrorType, _NewEnvType]:
        """
        Creates new container from ``RequiresContextResult`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_result_context(
          ...         RequiresContextResult.from_value(1),
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_result_context(
          ...         RequiresContextResult.from_failure(1),
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda deps: FutureResult.from_result(inner_value(deps)),
        )

    @classmethod
    def from_ioresult_context(
        cls,
        inner_value: ReaderIOResult[_NewValueType, _NewErrorType, _NewEnvType],
    ) -> ReaderFutureResult[_NewValueType, _NewErrorType, _NewEnvType]:
        """
        Creates new container from ``RequiresContextIOResult`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextIOResult
          >>> from returns.io import IOSuccess, IOFailure

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_ioresult_context(
          ...         RequiresContextIOResult.from_value(1),
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOSuccess(1)

          >>> assert anyio.run(
          ...     RequiresContextFutureResult.from_ioresult_context(
          ...         RequiresContextIOResult.from_failure(1),
          ...     ),
          ...     RequiresContextFutureResult.no_args,
          ... ) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda deps: FutureResult.from_ioresult(inner_value(deps)),
        )

    @classmethod
    def from_value(
        cls,
        inner_value: _FirstType,
    ) -> RequiresContextFutureResult[_FirstType, Any, NoDeps]:
        """
        Creates new container with successful ``FutureResult`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOSuccess

          >>> assert anyio.run(RequiresContextFutureResult.from_value(1)(
          ...    RequiresContextFutureResult.no_args,
          ... ).awaitable) == IOSuccess(1)

        """
        return RequiresContextFutureResult(
            lambda _: FutureResult.from_value(inner_value),
        )

    @classmethod
    def from_failure(
        cls,
        inner_value: _FirstType,
    ) -> RequiresContextFutureResult[Any, _FirstType, NoDeps]:
        """
        Creates new container with failed ``FutureResult`` as a unit value.

        .. code:: python

          >>> import anyio
          >>> from returns.context import RequiresContextFutureResult
          >>> from returns.io import IOFailure

          >>> assert anyio.run(RequiresContextFutureResult.from_failure(1)(
          ...     RequiresContextFutureResult.no_args,
          ... ).awaitable) == IOFailure(1)

        """
        return RequiresContextFutureResult(
            lambda _: FutureResult.from_failure(inner_value),
        )


# Aliases:

#: Alias for a popular case when ``Result`` has ``Exception`` as error type.
RequiresContextFutureResultE: TypeAlias = RequiresContextFutureResult[
    _ValueType_co,
    Exception,
    _EnvType_contra,
]

#: Sometimes `RequiresContextFutureResult` is too long to type.
ReaderFutureResult: TypeAlias = RequiresContextFutureResult

#: Alias to save you some typing. Uses ``Exception`` as error type.
ReaderFutureResultE: TypeAlias = RequiresContextFutureResult[
    _ValueType_co,
    Exception,
    _EnvType_contra,
]



from mypy.plugin import FunctionContext
from mypy.types import Type as MypyType
from mypy.types import get_proper_type

from returns.contrib.mypy._typeops.inference import PipelineInference


def analyze(ctx: FunctionContext) -> MypyType:
    """
    Helps to analyze ``flow`` function calls.

    By default, ``mypy`` cannot infer and check this function call:

    .. code:: python

      >>> from returns.pipeline import flow
      >>> assert flow(
      ...     1,
      ...     lambda x: x + 1,
      ...     lambda y: y / 2,
      ... ) == 1.0

    But, this plugin can!
    It knows all the types for all ``lambda`` functions in the pipeline.
    How?

    1. We use the first passed parameter as the first argument
       to the first passed function
    2. We use parameter + function to check the call and reveal
       types of current pipeline step
    3. We iterate through all passed function and use previous
       return type as a new parameter to call current function

    """
    if not ctx.arg_types[0]:
        return ctx.default_return_type
    if not ctx.arg_types[1]:  # We do require to pass `*functions` arg.
        ctx.api.fail('Too few arguments for "flow"', ctx.context)
        return ctx.default_return_type

    # We use custom argument type inference here,
    # because for some reason, `mypy` does not do it correctly.
    # It inferes `covariant` types incorrectly.
    real_arg_types = tuple(
        ctx.api.expr_checker.accept(arg)  # type: ignore
        for arg in ctx.args[1]
    )

    return PipelineInference(
        get_proper_type(ctx.arg_types[0][0]),
    ).from_callable_sequence(
        real_arg_types,
        ctx.arg_kinds[1],
        ctx,
    )



from collections.abc import Callable
from typing import TypeVar, overload

from returns.context import NoDeps
from returns.interfaces.failable import DiverseFailableN, SingleFailableN
from returns.methods.cond import internal_cond
from returns.primitives.hkt import Kinded, KindN

_ValueType = TypeVar('_ValueType')
_ErrorType = TypeVar('_ErrorType')

_DiverseFailableKind = TypeVar('_DiverseFailableKind', bound=DiverseFailableN)
_SingleFailableKind = TypeVar('_SingleFailableKind', bound=SingleFailableN)


@overload
def cond(
    container_type: type[_SingleFailableKind],
    success_value: _ValueType,
) -> Kinded[
    Callable[
        [bool],
        KindN[_SingleFailableKind, _ValueType, _ErrorType, NoDeps],
    ]
]: ...


@overload
def cond(
    container_type: type[_DiverseFailableKind],
    success_value: _ValueType,
    error_value: _ErrorType,
) -> Kinded[
    Callable[
        [bool],
        KindN[_DiverseFailableKind, _ValueType, _ErrorType, NoDeps],
    ]
]: ...


def cond(
    container_type: (type[_SingleFailableKind] | type[_DiverseFailableKind]),
    success_value: _ValueType,
    error_value: _ErrorType | None = None,
):
    """
    Reduce the boilerplate when choosing paths.

    Works with ``SingleFailableN`` (e.g. ``Maybe``)
    and ``DiverseFailableN`` (e.g. ``Result``).

    Example using ``cond`` with the ``Result`` container:

    .. code:: python

      >>> from returns.pointfree import cond
      >>> from returns.result import Failure, Result, Success

      >>> assert cond(Result, 'success', 'failure')(True) == Success('success')
      >>> assert cond(Result, 'success', 'failure')(False) == Failure('failure')

    Example using ``cond`` with the ``Maybe`` container:

    .. code:: python

      >>> from returns.maybe import Maybe, Some, Nothing

      >>> assert cond(Maybe, 10.0)(True) == Some(10.0)
      >>> assert cond(Maybe, 10.0)(False) == Nothing

    """

    def factory(is_success: bool):  # noqa: FBT001
        return internal_cond(
            container_type,
            is_success,
            success_value,
            error_value,
        )

    return factory



from copy import copy, deepcopy

import pytest

from returns.maybe import Nothing, Some, _Nothing  # noqa: PLC2701
from returns.primitives.exceptions import ImmutableStateError


def test_equals():
    """Ensures that ``.equals`` method works correctly."""
    inner_value = 1

    assert Some(inner_value).equals(Some(inner_value))
    assert Nothing.equals(Nothing)


def test_not_equals():
    """Ensures that ``.equals`` method works correctly."""
    assert not Some(1).equals(Nothing)
    assert not Some(1).equals(Some(0))
    assert not Nothing.equals(Some(1))


def test_equality():
    """Ensures that containers can be compared."""
    assert Nothing is Nothing  # noqa: PLR0124, WPS312
    assert Nothing == _Nothing() == _Nothing(None)
    assert Some(5) == Some(5)
    assert hash(Some(1))
    assert hash(Nothing)


def test_nonequality():
    """Ensures that containers are not compared to regular values."""
    assert Nothing is not None
    assert Nothing != None  # noqa: E711
    assert _Nothing(None) != None  # noqa: E711
    assert Some(5) != 5
    assert Some(3) is not Some(3)


def test_is_compare():
    """Ensures that `is` operator works correctly."""
    some_container = Some(1)

    assert Nothing.bind(lambda state: state) is Nothing
    assert some_container is not Some(1)


def test_immutability_failure():
    """Ensures that Failure container is immutable."""
    with pytest.raises(ImmutableStateError):
        Nothing._inner_state = 1  # noqa: SLF001

    with pytest.raises(ImmutableStateError):
        Nothing.missing = 2

    with pytest.raises(ImmutableStateError):
        del Nothing._inner_state  # type: ignore # noqa: SLF001, WPS420

    with pytest.raises(AttributeError):
        Nothing.missing  # type: ignore # noqa: B018


def test_immutability_success():
    """Ensures that Success container is immutable."""
    with pytest.raises(ImmutableStateError):
        Some(0)._inner_state = 1  # noqa: SLF001

    with pytest.raises(ImmutableStateError):
        Some(1).missing = 2

    with pytest.raises(ImmutableStateError):
        del Some(0)._inner_state  # type: ignore # noqa: SLF001, WPS420

    with pytest.raises(AttributeError):
        Some(1).missing  # type: ignore # noqa: B018


def test_success_immutable_copy():
    """Ensures that Success returns it self when passed to copy function."""
    some = Some(1)  # noqa: WPS110
    assert some is copy(some)


def test_success_immutable_deepcopy():
    """Ensures that Success returns it self when passed to deepcopy function."""
    some = Some(1)  # noqa: WPS110
    assert some is deepcopy(some)


def test_failure_immutable_copy():
    """Ensures that Failure returns it self when passed to copy function."""
    nothing = _Nothing()
    assert nothing is copy(nothing)


def test_failure_immutable_deepcopy():
    """Ensures that Failure returns it self when passed to deepcopy function."""
    nothing = _Nothing()
    assert nothing is deepcopy(nothing)
